""" functions to load genes, and identify transcripts containing de novos.
"""

import asyncio
import logging

from denovonear.transcript import Transcript
from denovonear.gencode import Gencode, Gene

from denovonear.ensembl_requester import (get_protein_seq_for_transcript,
    get_genomic_seq_for_transcript, get_cds_seq_for_transcript,
    get_cds_ranges_for_transcript, get_exon_ranges_for_transcript,
    get_genes_for_hgnc_id, get_transcript_ids_for_ensembl_gene_id,
    get_previous_symbol)

def cds_length(transcript):
    '''get length of CDS
    '''
    return transcript.get_coding_distance(transcript.get_cds_end())['pos'] + 1

async def construct_gene_object(ensembl, transcript_id):
    """ creates an Transcript object for a gene from ensembl databases
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        transcript_id: string for an Ensembl transcript ID
    
    Returns:
        a Transcript object, containing transcript coordinates and gene and
        transcript sequence.
    
    Raises:
        ValueError if CDS from genomic sequence given gene coordinates and CDS
        retrieved from Ensembl do not match.
    """
    tasks = [get_genomic_seq_for_transcript(ensembl, transcript_id, expand=10),
        get_cds_seq_for_transcript(ensembl, transcript_id),
        get_cds_ranges_for_transcript(ensembl, transcript_id),
        get_exon_ranges_for_transcript(ensembl, transcript_id)]
    (chrom, start, end, strand, genomic), cds_seq, cds, exons = await asyncio.gather(*tasks)
    
    # start a Transcript object with the locations and sequence
    transcript = Transcript(transcript_id, chrom, start, end, strand)
    transcript.set_exons(exons, cds)
    transcript.set_cds(cds)
    transcript.add_cds_sequence(cds_seq)
    transcript.add_genomic_sequence(genomic, offset=10)
    
    return transcript

def get_de_novos_in_transcript(transcript, de_novos):
    """ get the de novos within the coding sequence of a transcript
    
    Args:
        transcript: Transcript object, which defines the transcript coordinates
        de_novos: set of chromosome sequence positions for de novo events
    
    Returns:
        list of de novo positions found within the transcript
    """
    
    in_transcript = []
    for de_novo in de_novos:
        # we check if the de novo is within the transcript by converting the
        # chromosomal position to a CDS-based position. Variants outside the CDS
        # will raise an error, which we catch and pass on. It's better to do
        # this, rather than use the function in_coding_region(), since that
        # function does not allow for splice site variants.
        site = transcript.get_coding_distance(de_novo)
        within_cds = site['pos'] >= 0 and site['pos'] < cds_length(transcript)
        if within_cds and (transcript.in_coding_region(de_novo) or abs(site['offset']) < 9):
            in_transcript.append(de_novo)
    
    return in_transcript

def flatten(values):
    return [x for sub in values for x in sub]

async def get_transcript_ids(ensembl, symbol):
    """ gets transcript IDs for a gene.
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        symbol: HGNC symbol for gene
    
    Returns:
       list of transcript IDs for a given HGNC symbol.
    """
    
    genes = await get_genes_for_hgnc_id(ensembl, symbol)
    tasks = [get_transcript_ids_for_ensembl_gene_id(ensembl, x, symbol) for x in genes]
    transcript_ids = flatten(await asyncio.gather(*tasks))
    
    # sometimes we get HGNC symbols that do not match the ensembl rest version
    # that we are currentl using. We can look for earlier HGNC symbols for
    # the gene using the service at rest.genenames.org
    symbols = []
    if len(transcript_ids) == 0:
        symbols = await get_previous_symbol(ensembl, symbol)
        tasks = [get_genes_for_hgnc_id(ensembl, s) for s in symbols]
        genes += flatten(await asyncio.gather(*tasks))
        symbols = [symbol] + symbols
        
        tasks = [get_transcript_ids_for_ensembl_gene_id(ensembl, x, symbols) for x in genes]
        transcript_ids = flatten(await asyncio.gather(*tasks))
    
    return transcript_ids

async def load_gene(ensembl, gene_id):
    """ sort out all the necessary sequences and positions for a gene
    
    Args:
        ensembl: EnsemblRequest object to request data from ensembl
        gene_id: HGNC symbol for gene
        
    Returns:
        Gene object, containing info for each transcript
    """
    
    tx_ids = await get_transcript_ids(ensembl, gene_id)
    logging.info(f'found tx IDs for {gene_id}: {tx_ids}')
    
    gene = Gene(gene_id)
    for tx_id in tx_ids:
        logging.info(f'constructing: {tx_id}')
        try:
            tx = await construct_gene_object(ensembl, tx_id)
        except ValueError as e:
            logging.info(f'failed: {tx_id}')
            logging.info(e)
            continue
        gene.add_transcript(tx)
    
    if len(gene.transcripts) == 0:
        logging.info(f"{gene_id} lacks coding transcripts")
    
    return gene

def count_de_novos_per_transcript(transcripts, de_novos=[]):
    """ count de novos in transcripts for a gene.
    
    Args:
        transcripts: list of Transcript objects
        de_novos: list of de novo positions, so we can check they all fit in
            the gene transcript
        
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript IDs.
    """
    
    # count the de novos observed in all transcripts
    counts = {}
    for tx in transcripts:
        tx_id = tx.get_name()
        total = len(get_de_novos_in_transcript(tx, de_novos))
        counts[tx_id] = {'n': total, 'len': cds_length(tx)}
    
    return counts

def minimise_transcripts(transcripts, de_novos):
    """ get a set of minimal transcripts to contain all the de novos.
    
    We identify the minimal number of transcripts to contain all de novos. This
    allows for de novos on mutually exclusive transcripts. The transcripts are
    selected on the basis of containing the most number of de novos, while also
    being the longest possible transcript for the gene.
    
    Args:
        transcripts: list of Transcript objects
        de_novos: list of de novo positions
    
    Returns:
        dictionary of lengths and de novo counts, indexed by transcript ID for
        the set of minimal transcripts necessary to contain all de novos.
    """
    
    if len(de_novos) == 0:
        return {}
    
    counts = count_de_novos_per_transcript(transcripts, de_novos)
    
    # find the transcripts with the most de novos
    max_count = max( val["n"] for key, val in counts.items() )
    max_ids = [ key for key, val in counts.items() if val["n"] == max_count ]
    
    if max_count == 0:
        return {}
    
    # find the transcripts with the greatest length, and the most de novos
    max_length = max( counts[k]["len"] for k in max_ids )
    max_transcripts = {x: counts[x] for x in counts if x in max_ids and counts[x]['len'] == max_length}
    
    # find which de novos occur in the transcript with the most de novos
    best = [x for x in transcripts if x.get_name() == next(iter(max_transcripts))][0]
    denovos_in_gene = get_de_novos_in_transcript(best, de_novos)
    
    # trim the de novos to the ones not in the current transcript
    leftovers = [ x for x in de_novos if x not in denovos_in_gene ]
    
    # recursively return the transcripts in the current transcript, along
    # with transcripts for the reminaing de novos
    update = minimise_transcripts(transcripts, leftovers)
    max_transcripts.update(update)
    
    return max_transcripts



# cython: language_level=3, boundscheck=False
'''
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool

cdef extern from "tx.h":
    cdef cppclass Tx:
        Tx(string, string, int, int, char, string) except +
        Tx() except +
        
        void set_exons(vector[vector[int]], vector[vector[int]]) except +
        void set_cds(vector[vector[int]]) except +
        Region fix_cds_boundary(int) except +
        
        vector[Region] get_exons()
        vector[Region] get_cds()
        string get_name()
        string get_chrom()
        int get_start()
        int get_end()
        char get_strand()
        string get_type()
        int get_cds_start()
        int get_cds_end()
        
        bool is_exonic(int)
        int closest_exon_num(int)
        Region get_closest_exon(int)
        bool in_coding_region(int)
        CDS_coords to_closest_exon(int, Region)
        CDS_coords get_coding_distance(int) except +
        
        int get_position_on_chrom(int, int) except +
        int get_codon_number_for_cds_position(int)
        int get_position_within_codon(int)
        void add_cds_sequence(string)
        void add_genomic_sequence(string, int) except +
        string get_cds_sequence()
        string get_genomic_sequence()
        int get_genomic_offset()
        
        string reverse_complement(string)
        string get_centered_sequence(int, int) except +
        string get_codon_sequence(int) except +
        string get_seq_in_region(int, int) except +
        string translate(string) except +
        
        Codon get_codon_info(int) except +
        int get_boundary_distance(int) except +
        string consequence(int, string, string) except +
    
    cdef struct CDS_coords:
        int position
        int offset
    
    cdef struct Region:
        int start
        int end
    
    cdef struct Codon:
        int cds_pos
        string codon_seq
        int intra_codon
        int codon_number
        char initial_aa
        int offset

cdef class Transcript:
    cdef Tx *thisptr # hold a C++ instance which we're wrapping



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import math
import unittest

from denovonear.weights import get_distances, geomean, WeightedChoice, \
    analyse_de_novos, simulate_distribution

class TestSimulationsPy(unittest.TestCase):
    """ unit test the simulation functions
    """
    
    def setUp(self):
        """
        """
        
        # set up a range of possible positions, all with a uniform probability
        # of being selected
        self.choices = WeightedChoice()
        for x in range(1000):
            self.choices.add_choice(x, 0.0001)
        
        self.iterations = 100000
    
    def test_analyse_de_novos_dispersed(self):
        """ test analyse_de_novos() works correctly for dispersed de novos
        """
        
        # spread sites throughout a 1000 bp transcript
        positions = [100, 300, 600]
        distances = get_distances(positions)
        observed = geomean(distances)
        
        p_val = analyse_de_novos(self.choices, self.iterations, len(positions), observed)
        
        self.assertAlmostEqual(p_val, 0.635, places=2)
    
    def test_analyse_de_novos_clustered(self):
        """ test analyse_de_novos() works correctly for clustered de novos
        """
        
        # cluster sites within 20 bp in a 1000 bp transcript
        positions = [100, 110, 120]
        distances = get_distances(positions)
        observed = geomean(distances)
        
        p_val = analyse_de_novos(self.choices, 1000000, len(positions), observed)
        
        self.assertAlmostEqual(p_val, 0.002, places=3)
    
    def test_simulate_distribution(self):
        ''' check that simulate_distribution works correctly
        '''
        
        # repeated function calls should give different samples
        first = simulate_distribution(self.choices, iterations=5, de_novos_count=3)
        second = simulate_distribution(self.choices, iterations=5, de_novos_count=3)
        
        self.assertNotEqual(first, second)
        



# cython: language_level=3, boundscheck=False
'''
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
'''

from itertools import combinations

cdef class Transcript:
    def __cinit__(self, name, chrom, start, end, strand, 
            transcript_type='protein_coding', exons=None, cds=None, sequence=None, 
            offset=0):
        ''' construct a Transcript object
        
        Args:
            name: ID of the transcript
            start: position in bp at 5' edge of transcript (on + strand)
            end: position in bp at 3' edge of transcript (on + strand)
            exons: list of tuples defining start and end positions of exons
            cds: list of tuples defining start and end positions of CDS regions
            sequence: DNA sequence of genome region of the transcript.
            offset: how many base pairs the DNA sequence extends outwards
        '''
        
        name = name.encode('utf8')
        chrom = chrom.encode('utf8')
        transcript_type = transcript_type.encode('utf8')
        self.thisptr = new Tx(name, chrom, start, end, ord(strand), transcript_type)
        
        if exons is not None and cds is not None:
            self.set_exons(exons, cds)
            self.set_cds(cds)
            
            if sequence is not None:
                self.add_genomic_sequence(sequence, offset)
    
    def __dealloc__(self):
        del self.thisptr
    
    def __repr__(self):
        
        exons = [ (x['start'], x['end']) for x in self.get_exons() ]
        cds = [ (x['start'], x['end']) for x in self.get_cds() ]
        seq = self.get_genomic_sequence()
        
        if len(seq) > 40:
            seq = seq[:20] + '...[{} bp]...'.format(len(seq) - 40) + seq[-20:]
        
        if exons == []:
            exons = None
        
        if cds == []:
            cds = None
        
        if seq == '':
            seq = None
        else:
            seq = '"' + seq + '"'
        
        return f'Transcript(name="{self.get_name()}", chrom="{self.get_chrom()}", ' \
            f'start={self.get_start()}, end={self.get_end()}, strand="{self.get_strand()}", ' \
            f'transcript_type="{self.get_type()}", exons={exons}, cds={cds}, ' \
            f'sequence={seq}, offset={self.get_genomic_offset()})'
    
    def __str__(self):
        return self.__repr__()
    
    def __hash__(self):
        return hash((self.thisptr.get_chrom(), self.thisptr.get_start(),
            self.thisptr.get_end()))
    
    def __richcmp__(self, other, op):
        
        if op == 2:
            return self.__hash__() == other.__hash__()
        else:
            err_msg = "op {0} isn't implemented yet".format(op)
            raise NotImplementedError(err_msg)
    
    def set_exons(self, exon_ranges, cds_ranges):
        ''' add exon ranges
        
        Args:
            exon_ranges: a CDS position of the selected base.
        '''
        self.thisptr.set_exons(exon_ranges, cds_ranges)
    
    def get_overlaps(self, exon, regions):
        ''' find all regions which overlap a given region
        '''
        return [ i for i, x in enumerate(regions) if
            exon['start'] <= x['end'] and exon['end'] >= x['start'] ]
    
    def insert_region(self, coords, region):
        ''' include a region into a list of regions
        
        To include a region, we have to check which pre-existing regions the new
        region overlaps, so any overlaps can be merged into a single region.
        
        Args:
            coords: list of {start: X, end: Y} dictionaries
            region: dict of {'start': X, 'end': Y} positions
        '''
        indices = self.get_overlaps(region, coords)
        overlaps = [ coords[i] for i in indices ]
        start = min( x['start'] for x in overlaps + [region] )
        end = max( x['end'] for x in overlaps + [region] )
        
        for i in sorted(indices, reverse=True):
            del coords[i]
        
        return coords + [{'start': start, 'end': end}]
    
    def merge_coordinates(self, first, second):
        ''' merge two sets of coordinates, to get the union of regions
        
        This uses an inefficient approach, looping over and over, but we won't
        need to perform this often.
        
        Args:
            first: list of {'start': x, 'end': y} dictionaries for first transcript
            second: list of {'start': x, 'end': y} dictionaries for second transcript
        
        Returns:
            list of [start, end] lists, sorted by position.
        '''
        coords = []
        for a, b in combinations(first + second, 2):
            if a['start'] <= b['end'] and a['end'] >= b['start']:
                region = {'start': min(a['start'], b['start']),
                    'end': max(a['end'], b['end'])}
                a, b = region, region
            
            coords = self.insert_region(coords, a)
            coords = self.insert_region(coords, b)
        
        return [ (x['start'], x['end']) for x in sorted(coords, key=lambda x: x['start']) ]
    
    def merge_genomic_seq(self, other):
        ''' merge the genomic sequence from two transcripts
        
        We have two transcripts A, and B. We need to get the contiguous sequence
        from the start of the first transcript on the chromosome to the end of
        the second transcript. The transcripts may or may not overlap. There are
        three scenarios we need to account for:
        
        
        overlap without   A   =================
        enveloping                           ================= B
        
        
            overlap       A   ==============================
        and envelop               ==================  B
        
        
        no overlap        A  ===============
                                               ===============  B
        
        I've called the transcript whose sequence is first along the chromosome
        as 'lead', and the transcript whose sequence is last as 'lag', and the
        converse as 'not_lead', and 'not_lag'. Note that in the envelope case,
        the lead transcript can also be the lag transcript.
        '''
        
        # make sure that the surrounding sequence is the same length in both
        # transcripts.
        # TODO: this could be worked around, by figuring the minimum offset length,
        # TODO: then trimming the respective DNA offset sequences to that length.
        assert self.get_genomic_offset() == other.get_genomic_offset()
        
        # figure out which transcripts hold the leading and lagging sections
        lead, not_lead = self, other
        if self.get_start() > other.get_start():
            lead, not_lead = other, self
        
        lag, not_lag = self, other
        if self.get_end() < other.get_end():
            lag, not_lag = other, self
        
        lead_offset = lead.get_genomic_offset()
        lead_gdna = lead.get_genomic_sequence()
        initial = lead_gdna[:not_lead.get_start() - lead.get_start() + lead_offset]
        
        if self.get_start() <= other.get_end() and self.get_end() >= other.get_start():
            intersect_start = not_lead.get_start() - lead.get_start() + lead_offset
            intersect_end = not_lag.get_end() - lead.get_start() + lead_offset
            intersect = lead_gdna[intersect_start:intersect_end]
        else:
            intersect = 'N' * (lag.get_start() - lead.get_end() - lead_offset * 2)
        
        lag_offset = lag.get_genomic_offset()
        lag_gdna = lag.get_genomic_sequence()
        
        # some transcripts overlap, but some do not. We need to find the position
        # where the lagging transcript takes over, which is either at the end of
        # not lagging transcript, or the start of the lagging transcript,
        # whichever is higher
        pos = max(not_lag.get_end(), lag.get_start())
        final = lag_gdna[pos - lag.get_start() + lead_offset:]
        
        return initial + intersect + final
    
    def __add__(self, other):
        """ combine the coding sequences of two Transcript objects
        
        When we determine the sites for sampling, occasioally we want to
        use sites from multiple alternative transcripts. We determine the sites
        for each transcript in turn, but mask the sites that have been collected
        in the preceeding transcripts. In order to be able to mask all previous
        trabnscripts, we need to combine the coding sequence of the transcripts
        as we move through them. This function performs the union of coding
        sequence regions between different transcripts.
        
        We do this outside of the c++ class, so as to be able to set up a
        Transcript object correctly.
        
        Args:
            other: a transcript to be combined with the current object.
        
        Returns:
            an altered instance of the class, where the coding sequence regions
            are the union of the coding regions of two Transcript objects. This
            disrupts the ability to get meaningingful sequence from the object,
            so don't try to extract sequence from the returned object.
        """
        
        # if we try transcript + None or None + transcript, return the original
        # transcript, rather than raising an error.
        if other is None:
            return self
        if self is None:
            return other
        
        altered = Transcript('{}:{}'.format(self.get_name(), other.get_name()),
            self.get_chrom(), min(self.get_start(), other.get_start()),
            max(self.get_end(), other.get_end()), self.get_strand(), self.get_type())
        
        exons = self.merge_coordinates(self.get_exons(), other.get_exons())
        cds = self.merge_coordinates(self.get_cds(), other.get_cds())
        
        altered.set_exons(exons, cds)
        altered.set_cds(cds)
        
        if self.get_genomic_sequence() != "":
            altered.add_genomic_sequence(self.merge_genomic_seq(other), self.get_genomic_offset())
        
        return altered
    
    def set_cds(self, cds_ranges):
        ''' set CDS ranges
        
        Args:
            cds_ranges: a CDS position of the selected base.
        '''
        
        self.thisptr.set_cds(cds_ranges)
    
    def get_genomic_offset(self):
        return self.thisptr.get_genomic_offset()
    def get_exons(self):
        return self.thisptr.get_exons()
    def get_cds(self):
        return self.thisptr.get_cds()
    def get_name(self):
        return self.thisptr.get_name().decode('utf8')
    def get_chrom(self):
        return self.thisptr.get_chrom().decode('utf8')
    def get_type(self):
        return self.thisptr.get_type().decode('utf8')
    def get_start(self):
        return self.thisptr.get_start()
    def get_end(self):
        return self.thisptr.get_end()
    def get_strand(self):
        return chr(self.thisptr.get_strand())
    def get_cds_start(self):
        return self.thisptr.get_cds_start()
    def get_cds_end(self):
        return self.thisptr.get_cds_end()
    def fix_cds_boundary(self, pos):
        return self.thisptr.fix_cds_boundary(pos)
    
    def in_exons(self, position):
        ''' check if a site lies within the exon ranges
        
        Args:
            position: an integer-based chromosome position.
        '''
        
        return self.thisptr.is_exonic(position)
    
    def get_closest_exon(self, position):
        ''' finds the positions of the exon closest to a position
        '''
        return self.thisptr.get_closest_exon(position)
    
    def in_coding_region(self, position):
        return self.thisptr.in_coding_region(position)
    
    def get_coding_distance(self, pos):
        ''' get distance to CDS start (and intronic offset)
        '''
        coords = self.thisptr.get_coding_distance(pos)
        
        return {'pos': coords.position, 'offset': coords.offset}
    
    def get_position_on_chrom(self, pos, offset=0):
        return self.thisptr.get_position_on_chrom(pos, offset)
    
    def get_codon_number_for_cds_position(self, pos):
        return self.thisptr.get_codon_number_for_cds_position(pos)
    
    def get_position_within_codon(self, pos):
        return self.thisptr.get_position_within_codon(pos)
    
    def add_cds_sequence(self, text):
        self.thisptr.add_cds_sequence(text.encode('utf8'))
        
    def get_cds_sequence(self):
        return self.thisptr.get_cds_sequence().decode('utf8')
    
    def add_genomic_sequence(self, text, offset=0):
        self.thisptr.add_genomic_sequence(text.encode('utf8'), offset)
    
    def get_genomic_sequence(self):
        return self.thisptr.get_genomic_sequence().decode('utf8')
    
    def reverse_complement(self, text):
        return self.thisptr.reverse_complement(text).decode('utf8')
    
    def get_centered_sequence(self, pos, length=3):
        return self.thisptr.get_centered_sequence(pos, length).decode('utf8')
    
    def get_codon_sequence(self, pos):
        return self.thisptr.get_codon_sequence(pos).decode('utf8')
    
    def translate(self, text):
        return self.thisptr.translate(text.encode('utf8')).decode('utf8')
    
    def get_codon_info(self, pos):
        codon = dict(self.thisptr.get_codon_info(pos))
        
        if codon['codon_number'] == -9999999:
            codon['codon_number'] = None
            codon['intra_codon'] = None
            codon['codon_seq'] = None
            codon['initial_aa'] = None
        
        if codon['codon_seq'] is not None:
            codon['codon_seq'] = codon['codon_seq'].decode('utf8')
        
        if codon['initial_aa'] is not None:
            codon['initial_aa'] = chr(codon['initial_aa'])
        
        return codon
    
    def get_boundary_distance(self, pos):
        return self.thisptr.get_boundary_distance(pos)
    
    def consequence(self, pos, ref, alt):
        cq = self.thisptr.consequence(pos, ref.encode('utf8'), alt.encode('utf8'))
        return cq.decode('utf8')



""" loads a file containing known de novo mutations
"""

from __future__ import print_function, division

missense = ["missense_variant", "stop_lost", "inframe_deletion",
    "inframe_insertion", "coding_sequence_variant", "protein_altering_variant"]

lof = ["stop_gained", "splice_acceptor_variant",
    "splice_donor_variant", "frameshift_variant", "initiator_codon_variant",
    "start_lost", "conserved_exon_terminus_variant"]
    
synonymous = ["synonymous_variant"]

def load_de_novos(path, exclude_indels=True):
    """ load mutations into dict indexed by HGNC ID.
    
    Args:
        path: path to file containing de novo data. This should have five tab-
            separated columns e.g.
            hgnc  chr  pos      consequence         var_type
            CTC1  17   8139190  missense_variant    snv
            CTC1  17   8139191  frameshift_variant  indel
        exclude_indels: True/False for whether we want to exclude indels. If we
            are testing clustering of de novos, we can only use SNVs, but if we
            are determining mutation rates for a gene, then we include indels,
            in order to identify the transcripts that the de novos lie within.
    
    Returns:
        dictionary of missense and lof counts for each gene, indexed by HGNC symbols
    """
    
    genes = {}
    with open(path, "r") as handle:
        header = handle.readline().strip().split("\t")
        for line in handle:
            
            line = line.rstrip().split("\t")
            gene = line[0]
            position = int(line[2]) - 1
            consequence = line[3]
            var_type = line[4]
            
            # ignore indels (some splice_acceptor_variants (in the
            # functional_consequences) are indels
            if exclude_indels and "indel" in var_type.lower():
                continue
            
            # trim out variants that are missing data
            if gene == "" or gene == "." or position == "NA":
                continue
            
            if gene not in genes:
                genes[gene] = {"missense": [], "nonsense": []}
            
            if consequence in missense:
                genes[gene]["missense"].append(position)
            elif consequence in lof:
                genes[gene]["nonsense"].append(position)
        
    return genes
        
        
        
    
    
    
    



"""
Copyright (c) 2015 Genome Research Ltd.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import asyncio
import unittest
import tempfile
from pathlib import Path

from denovonear.load_gene import cds_length, construct_gene_object, \
    get_de_novos_in_transcript, get_transcript_ids, load_gene, \
    count_de_novos_per_transcript, minimise_transcripts
from denovonear.transcript import Transcript
from denovonear.rate_limiter import RateLimiter

async def call(func, *args, **kwargs):
    ''' call ensembl rest API function
    '''
    async with RateLimiter(15) as ensembl:
        return await func(ensembl, *args, **kwargs)

def _run(func, *args, **kwargs):
    return asyncio.get_event_loop().run_until_complete(call(func, *args, **kwargs))

class TestLoadGenePy(unittest.TestCase):
    """ unit test functions to load genes
    """
    
    def set_transcript(self):
        """ construct a transcript for a known gene
        """
        
        exon_ranges=[(120933859, 120934019), (120934219, 120934356),
            (120935876, 120936296)]
        cds_ranges=[(120934225, 120934356), (120935876, 120936013)]
        
        expected = Transcript("ENST00000242577", '12', 120933859, 120936296, "+")
        expected.set_exons(exon_ranges, cds_ranges)
        expected.set_cds(cds_ranges)
        
        cds = "ATGTGCGACCGAAAGGCCGTGATCAAAAATGCGGACATGTCGGAAGAGATGCAACAGGACTC" \
            "GGTGGAGTGCGCTACTCAGGCGCTGGAGAAATACAACATAGAGAAGGACATTGCGGCTCATATC" \
            "AAGAAGGAATTTGACAAGAAGTACAATCCCACCTGGCATTGCATCGTGGGGAGGAACTTCGGTA" \
            "GTTATGTGACACATGAAACCAAACACTTCATCTACTTCTACCTGGGCCAAGTGGCCATTCTTCT" \
            "GTTCAAATCTGGTTAA"
        
        genomic = "GGGCGGGGCCGGCGGGAGCAGGGCGGGGCCTGAGCACTAGGCGGCGGCGGCTGGCGTGGG" \
            "GCTGCTTAGATGCGCCACGGTTTCGGTAGCGACGGTATCTCTAGCCGGGCCTGAGCTGTGCTAGCA" \
            "CCTCCCCCAGGAGACCGTTGCAGTCGGCCAGCCCCCTTCTCCACGGTGAGAAACTCGGGGGGCCAG" \
            "GGGGTGTCCTCGCTGCCTTATTTCGCCCCACTCCGGACTTAGCCCTCCGCGTAGCCCGCGCTTCCT" \
            "GAGAAGTGGGGTGGGGGGCGTCGTCCCGTGGTGGCGCCGGCCGGGGTGGGGGCAGTTAGTGCCTGG" \
            "GGGGCGCGGCCCAACTCAACCCCTTACCCCAGGCCTTGCCCACTAGGTAACCATGTGCGACCGAAA" \
            "GGCCGTGATCAAAAATGCGGACATGTCGGAAGAGATGCAACAGGACTCGGTGGAGTGCGCTACTCA" \
            "GGCGCTGGAGAAATACAACATAGAGAAGGACATTGCGGCTCATATCAAGAAGGTGAGGATGGGCGC" \
            "GGGGGCCGATACGCAGCCGGGAGCAGGGGGTTCCTTCCCCCCGATCCTGCTTTCCTAAGGGCGCCT" \
            "GACAGGTCCCGGGAATACTGCTGGCGGCTTGGGGCGTAGAAGCTTCCAGAAAGGACGCAGATGCAT" \
            "TTTGCGCTCCTGTGGAGAAGACCAGACCCCCGGCGTCCGAAGTTTTTTTTTTTTTTTTTTTAATTA" \
            "CCCAGCTCCGCGGGGGGAAAGCGCCACCTAGCAACGGTATCTAAGATCAGGGAGCAGCGGTTCCCC" \
            "CTTCTGTGTGGTTCCTGCGCCGAGGATCCATCTGGGTGTTCCGGAGGGGGGAGCTGCGTGGGTGTT" \
            "TCCAGCCGGGCCGGGAGGAGATCTTGCCAGCCTTCCAGTGGGGAGTTGAGGGAAGGTGGTGGGTGG" \
            "TGGCGGGGCTGGGGGCTGGGGTAGGGGCTTGGTAAATGGCAGTCTAGAAAGCCGGCAGGACTGCCA" \
            "ACTTCTCGAGCAGTGTTTGCTGGAAGGGAAGAAAGCTGGCAGCCTAAGCCGTGGGAGGGTTCCAGT" \
            "CGAGAATGGGAAGATGAAAGACTTCAGATGGAACAGAAATAAATGCCTTTTTTGACAAACGCAGCA" \
            "GTGCGTGCCTCTAGCTTGCAAGAGCGTTACTCCCCTTCATAGCTTTAAAAGGTTTTCGCACTGCGT" \
            "GCAGTTAGAGTAGCTAAATCTTGTGTGACGCTCCACAAACACTTGTAAGAATTTTGCAGAGAAAGA" \
            "TAACCGTTGCCACCCAATGCCCCCCACAGGCATTCTACTCCCCAGTACCTCTTAGGGTGGGAGAAA" \
            "TGGTGAAGAGTTGTTCCTACAACTTGCTAACCTAGTGGACAGGGTAGTAGATTAGCATCATCCGGA" \
            "TAGATGTGAAGAGGACGGCTGTTTGGATAATAATTAAGGATAAAATTTGGCCAGTTGACAGATTCT" \
            "GTTTCCAGCAGTTTTTACAGCAACAGTGGAGTGCTTCAGTATTGTGTTCCTGTAAATTTAATTTTG" \
            "ATCCGCAATCATTTGGTATACAATGCTGTTTGAAGTTTTGTCCTATTGGAAAAGTCTTGTGTTGCA" \
            "GGGGTGCAGTTAAGATCTTTGTGATGAGGAATGGGATGGGCTAATTTTTTGCCGTTTTCTTGGAAT" \
            "TGGGGGCATGGCAAATACAGTAGGGTAGTTTAGTTCTCTACACAGAACATGATAAACTACACCTGT" \
            "TGATGTCACCGTCTGTCAATGAATATTATAGAAGGTATGAAGGTGTAATTACCATAATAACAAAAC" \
            "ACCCTGTCTTTAGGGCTGACCTTTCGTCCTTTGACCTCCTCAGCCTCCATTCCCATCTTCGCTCAG" \
            "ACTGCAAGTATGTTTGTATTAATGTACTATGTAGGCGGCTTGGAGCTGGGGAACATTCTTTCATTC" \
            "TAAGAATTTGCAGATGCTGACGTTCCTCCTTTCTGCCCCTACAGGCTCTGGCTTATCCAAGAGGCA" \
            "AACACTGACCTCTGGTAATTAAAATCCTAGTTCTTTTCTTTTGTCTTTTCCAGGAATTTGACAAGA" \
            "AGTACAATCCCACCTGGCATTGCATCGTGGGGAGGAACTTCGGTAGTTATGTGACACATGAAACCA" \
            "AACACTTCATCTACTTCTACCTGGGCCAAGTGGCCATTCTTCTGTTCAAATCTGGTTAAAAGCATG" \
            "GACTGTGCCACACACCCAGTGATCCATCCAAAAACAAGGACTGCAGCCTAAATTCCAAATACCAGA" \
            "GACTGAAATTTTCAGCCTTGCTAAGGGAACATCTCGATGTTTGAACCTTTGTTGTGTTTTGTACAG" \
            "GGCATTCTCTGTACTAGTTTGTCGTGGTTATAAAACAATTAGCAGAATAGCCTACATTTGTATTTA" \
            "TTTTCTATTCCATACTTCTGCCCACGTTGTTTTCTCTCAAAATCCATTCCTTTAAAAAATAAATCT" \
            "GATGCAGATGTGTATGTGTGTG"
        
        expected.add_cds_sequence(cds)
        expected.add_genomic_sequence(genomic, offset=10)
        
        return expected
    
    def _construct_transcript(self, tx_id):
        ''' helper function to prepare a Transcript given a transcript ID
        '''
        return _run(construct_gene_object, tx_id)
    
    def test_cds_length(self):
        ''' check we get the correct CDS length for a transcript
        '''
        tx = self.set_transcript()
        self.assertEqual(cds_length(tx), 270)
        
        tx = self._construct_transcript('ENST00000242577')
        self.assertEqual(cds_length(tx), 270)
        
        tx = self._construct_transcript('ENST00000392509')
        self.assertEqual(cds_length(tx), 270)
        
        tx = self._construct_transcript('ENST00000549649')
        self.assertEqual(cds_length(tx), 126)
    
    def test_construct_gene_object(self):
        """
        """
        
        transcript_id = "ENST00000242577"
        transcript = _run(construct_gene_object, transcript_id)
        
        expected = self.set_transcript()
        
        self.assertEqual(transcript, expected)
        self.assertEqual(transcript.get_genomic_sequence(), expected.get_genomic_sequence())
        self.assertEqual(transcript.get_cds_sequence(), expected.get_cds_sequence())
    
    def test_get_de_novos_in_transcript(self):
        """ test that we can identify de novos within the CDS of a transcript
        """
        
        exon_ranges = [(10, 20), (30, 40), (90, 100)]
        cds_ranges = [(30, 40), (90, 95)]
        
        # define a simple transcript
        tx = Transcript("test1", '1', 10, 100, "+")
        tx.set_exons(exon_ranges, cds_ranges)
        tx.set_cds(cds_ranges)
        
        # check that only the site in the CDS is returned
        sites = [15, 35, 100]
        self.assertEqual(get_de_novos_in_transcript(tx, sites), [35])
        
        # check that we can return multiple sites in the CDS
        sites = [15, 35, 90]
        self.assertEqual(get_de_novos_in_transcript(tx, sites), [35, 90])
        
        # check if we pass in an empty list, we get one back
        self.assertEqual(get_de_novos_in_transcript(tx, []), [])
    
    def test_get_transcript_ids(self):
        """ check that we get the correct transcript IDs for a gene
        """
        
        hgnc = "DYNLL1"
        lengths = _run(get_transcript_ids, hgnc)
        
        expected = ['ENST00000548342', 'ENST00000549989', 'ENST00000392509',
                    'ENST00000392508', 'ENST00000242577', 'ENST00000550845', 
                    'ENST00000550178', 'ENST00000548214', 'ENST00000552870',
                    'ENST00000549649']
        
        self.assertEqual(set(lengths), set(expected))
    
    def test_load_gene(self):
        """ check that we correctly load the suitable transcripts for a gene
        """
        
        # define a hgnc symbol to load transcript for, and de novo sites to
        # check against
        hgnc = "DYNLL1"
        sites = [120934226, 120936012]
        gene = _run(load_gene, hgnc)
        counts = minimise_transcripts(gene.transcripts, sites)
        
        # define the expected transcript, and make sure that it is in the list
        # of suitable transcripts. There can be multiple transcripts return if
        # more than one transcript of the maximal length includes all de novos.
        expected = self.set_transcript()
        self.assertIn(expected.get_name(), counts)
        
        # and make sure if none of the de novos fall in a suitable transcript,
        # then we get an empty dict.
        sites = [100, 200]
        gene = _run(load_gene, hgnc)
        transcripts = gene.transcripts
        counts = minimise_transcripts(transcripts, sites)
        self.assertEqual(counts, {})
    
    def test_count_de_novos_per_transcript(self):
        """ test that we count de novos in transcripts correctly
        """
        sites = [120934226, 120936012]
        
        expected = {'ENST00000549649': {'len': 126, 'n': 1},
            'ENST00000548214': {'len': 204, 'n': 1},
            'ENST00000242577': {'len': 270, 'n': 2},
            'ENST00000392508': {'len': 270, 'n': 2},
            'ENST00000392509': {'len': 270, 'n': 2},
            'ENST00000549989': {'len': 270, 'n': 2},
            'ENST00000550178': {'len': 204, 'n': 1},
            'ENST00000552870': {'len': 144, 'n': 1},
            'ENST00000550845': {'len': 204, 'n': 1},
            'ENST00000548342': {'len': 270, 'n': 2}}
        
        transcripts = [self._construct_transcript(x) for x in expected]
        
        counts = count_de_novos_per_transcript(transcripts, sites)
        self.assertEqual(counts, expected)
        
        # TODO: add test case for error from gene where no protein coding
        # TODO: transcript is available
        
    def test_minimise_transcripts(self):
        """ test that minimise_transcripts() works correctly
        """
        
        # run through a test case for a single gene
        sites = [120934226, 120936012]
        expected = {'ENST00000242577': {'len': 270, 'n': 2},
            'ENST00000392508': {'len': 270, 'n': 2},
            'ENST00000392509': {'len': 270, 'n': 2},
            'ENST00000549989': {'len': 270, 'n': 2},
            'ENST00000548342': {'len': 270, 'n': 2}}
        
        transcripts = [self._construct_transcript(x) for x in expected]
        counts = minimise_transcripts(transcripts, sites)
        
        self.assertEqual(counts, expected)
        
        # check that when we don't have any de novos, we return an empty list
        self.assertEqual(minimise_transcripts(transcripts, []), {})
        
        # check that when none of the de novos are in a transcript, we return
        # an empty list.
        self.assertEqual(minimise_transcripts(transcripts, [100]), {})




import math
import sys

def include_frameshift_rates(path):
    """ add per-gene frameshift mutation rates to the output file
    
    We make a crude estimate of the frameshift mutation rate by assessing the
    total frameshift burden (across all genes in a dataset) as 1.25 times the
    nonsense burden. Then apportion the frameshift burden proportional to each
    genes CDS length.
    
    This as as per Nature Genetics 46:944-950 (2014) doi:10.1038/ng.3050.
    Note that this is an approximation, and doesn't allow for divergence due to
    the base compostion of the CDS.
    
    Args:
        path: path to the output mutation rates (for the nonsense, missense etc)
    """
    
    if path == sys.stdout:
        raise ValueError("can't get frameshift rates when writing to standard out")
    
    with open(path) as handle:
        lines = [ x.strip().split('\t') for x in handle ]
    
    nonsense = sum([ 10**(float(x[4])) for x in lines[1:] if x[4] != 'NA' ])
    length = sum([ int(x[2]) for x in lines[1:] if x[2] != 'NA' ])
    
    # add the frameshift rates to each line in turn, while writing the output
    # back to the output path
    frameshift_sum = nonsense * 1.25
    with open(path, "w") as handle:
        for line in lines:
            if line[0] == "transcript_id":
                line.append("frameshift_rate")
            elif line[4] == 'NA':
                line.append('NA')
            else:
                # estimate the frameshift rate for the gene
                frameshift = (float(line[2])/length) * frameshift_sum
                frameshift = math.log10(frameshift)
                line.append(str(frameshift))
            
            line = "\t".join(line) +"\n"
            handle.write(line)



""" splits de novo clustering analysis across multiple jobs on a LSF cluster.
"""

import os
import subprocess
import time
import random
import argparse
import tempfile

def get_options():
    """ get the command line options
    """
    
    parser = argparse.ArgumentParser(description="Script to batch process de"
        "novo clustering.")
    parser.add_argument("--in", dest="input", required=True, help="Path to"
        "file listing known mutations in genes. See example file in data folder"
        "for format.")
    parser.add_argument("--temp-dir", required=True, help="path to hold intermediate files")
    parser.add_argument("--out", required=True, help="Path to output file.")
    
    args = parser.parse_args()
    
    return args

def count_missense_per_gene(lines):
    """ count the number of missense variants in each gene.
    """
    
    counts = {}
    for x in lines:
        x = x.split("\t")
        gene = x[0]
        consequence = x[3]
        
        if gene not in counts:
            counts[gene] = 0
        
        if consequence != "missense_variant":
            continue
        
        counts[gene] += 1
    
    return counts

def split_denovos(denovo_path, temp_dir):
    """ split de novos from an input file into files, one for each gene
    """
    
    # open the de novos, drop the header line, then sort them (which will be by
    # HGNC symbol, as the first element of each line)
    with open(denovo_path, "r") as handle:
        lines = handle.readlines()
        header = lines.pop(0)
    
    basename = os.path.basename(denovo_path)
    
    # only include genes with 2+ missense SNVs
    counts = count_missense_per_gene(lines)
    counts = dict((k, v) for k, v in counts.items() if v > 1 )
    
    genes = set([])
    for line in sorted(lines):
        gene = line.split("\t")[0]
        
        # open a new output file whenever we hit a different gene
        if gene not in genes and gene in counts:
            genes.add(gene)
            path = os.path.join(temp_dir, "tmp.{}.txt".format(len(genes)))
            output = open(path, "w")
            output.write(header)
        
        if gene in counts:
            output.write(line)
    
    return len(genes)

def is_number(string):
    """ check whether a string can be converted to a number
    
    Args:
        string: value as a string, could be a number
        
    Returns:
        True/False for whether the value can be converted to a number
    """
    
    try:
        number = float(string)
    except ValueError:
        return False
    
    return True

def get_random_string():
    """ make a random string, which we can use for bsub job IDs, so that
    different jobs do not have the same job IDs.
    """
    
    # set up a random string to associate with the run
    hash_string = "%8x" % random.getrandbits(32)
    hash_string = hash_string.strip()
    
    # done't allow the random strings to be equivalent to a number, since
    # the LSF cluster interprets those differently from letter-containing
    # strings
    while is_number(hash_string):
        hash_string = "%8x" % random.getrandbits(32)
        hash_string = hash_string.strip()
    
    return hash_string

def submit_bsub_job(command, job_id=None, dependent_id=None, memory=None, requeue_code=None, logfile=None):
    """ construct a bsub job submission command
    
    Args:
        command: list of strings that forma unix command
        job_id: string for job ID for submission
        dependent_id: job ID, or list of job IDs which the current command needs
            to have finished before the current command will start. Note that
            the list can be empty, in which case there are no dependencies.
        memory: minimum memory requirements (in megabytes)
    
    Returns:
        nothing
    """
    
    if job_id is None:
        job_id = get_random_string()
    
    job = "-J \"{0}\"".format(job_id)
    
    mem = ""
    if memory is not None:
        mem = "-R 'select[mem>{0}] rusage[mem={0}]' -M {0}".format(memory)
    
    requeue = ""
    if requeue_code is not None:
        requeue = "-Q 'EXCLUDE({0})'".format(requeue_code)
    
    dependent = ""
    if dependent_id is not None:
        if type(dependent_id) == list:
            dependent_id = " && ".join(dependent_id)
        dependent = "-w '{0}'".format(dependent_id)
    
    log = "bjob_output.txt"
    if logfile is not None:
        log = logfile
    
    preamble = ["bsub", job, dependent, requeue, "-q", "normal", "-o", log, mem]
    command = ["bash", "-c", "\""] + command + ["\""]
    
    command = " ".join(preamble + command)
    subprocess.call(command, shell=True)

def batch_process(de_novo_path, temp_dir, output_path):
    """ sets up a lsf job array
    """
    
    temp_dir = tempfile.mkdtemp(dir=temp_dir)
    count = split_denovos(de_novo_path, temp_dir)
    
    # set up run parameters
    job_name = "denovonear"
    job_id = "{0}[1-{1}]%20".format(job_name, count)
    
    basename = os.path.basename(de_novo_path)
    infile = os.path.join(temp_dir, "tmp.\$LSB_JOBINDEX\.txt")
    outfile = os.path.join(temp_dir, "tmp.\$LSB_JOBINDEX\.output")
    
    command = ["denovonear", "cluster",
        "--in", infile,
        "--out", outfile]
    submit_bsub_job(command, job_id, memory=3500, requeue_code=134, logfile="clustering.bjob")
    time.sleep(2)
    
    # merge the array output after the array finishes
    merge_id = "{}_merge".format(job_name)
    command = ["head", "-n", "1", os.path.join(temp_dir, "tmp.1.output"), ">", output_path, \
        "; tail", "-q", "-n", "+2", os.path.join(temp_dir, "tmp.*.output"), "|", "sort", ">>", output_path]
    submit_bsub_job(command, merge_id, memory=100, dependent_id=job_id, logfile="clustering.bjob")
    time.sleep(2)
    
    # submit a cleanup job to the cluster
    submit_bsub_job(["rm", "-r", temp_dir], job_id="{}_cleanup".format(job_name), \
        memory=100, dependent_id=merge_id, logfile="clustering.bjob")
    
def main():
    args = get_options()
    
    batch_process(args.input, args.temp_dir, args.out)

if __name__ == '__main__':
    main()




import asyncio
from pathlib import Path
import unittest
import tempfile

from denovonear.rate_limiter import RateLimiter
from denovonear.load_gene import construct_gene_object
from denovonear.gencode import Gencode, _parse_gtfline, _open_gencode

async def call(func, *args, **kwargs):
    ''' call ensembl rest API function
    '''
    async with RateLimiter(15) as ensembl:
        return await func(ensembl, *args, **kwargs)

def _run(func, *args, **kwargs):
    return asyncio.get_event_loop().run_until_complete(call(func, *args, **kwargs))

def write_gtf(path, lines):
    with open(path, 'wt') as output:
        for line in lines:
            output.write(line)

def make_fasta(path, chroms):
    with open(path, 'wt') as output:
        for chrom in chroms:
            output.write(f'>{chrom}\n')
            lines = ['A' * 50 + '\n'] * 50
            output.writelines(lines)

class TestGencode(unittest.TestCase):
    
    def setUp(self):
        ''' set path to folder with test data
        '''
        self.folder = Path(__file__).parent.parent /  "data"
        self.gtf_path = self.folder / 'example.grch38.gtf'
        self.fasta_path = self.folder / 'example.grch38.fa'
    
    def test_gencode_opens(self):
        ''' test we can open a gencode object
        '''
        gencode = Gencode(self.gtf_path, self.fasta_path)
        self.assertEqual(len(gencode), 1)
        gene = gencode['OR4F5']
        with self.assertRaises(KeyError):
            gencode['ZZZZZZZ']
        
    def test_gencode_matches_ensembl(self):
        ''' test thaqt gencode data matches ensembl
        '''
        gencode = Gencode(self.gtf_path, self.fasta_path)
        for gencode_tx in gencode['OR4F5'].transcripts:
            # get the transcript ID (but trim the version number)
            tx_id = gencode_tx.get_name().split('.')[0]
            ensembl_tx = _run(construct_gene_object, tx_id)
            self.assertEqual(gencode_tx.get_exons(), ensembl_tx.get_exons())
            self.assertEqual(gencode_tx.get_cds(), ensembl_tx.get_cds())
            self.assertEqual(gencode_tx.get_cds_sequence(), ensembl_tx.get_cds_sequence())
    
    def test_gencode_in_region(self):
        ''' test that in_region pulls out the correct genes
        '''
        lines = '##format: gtf\n' \
                'chr1\tHAVANA\tgene\t10\t20\t.\t-\t.\tgene_name "TEST1";\n' \
                'chr1\tHAVANA\ttranscript\t10\t20\t.\t-\t.\ttranscript_id "ENST_A";gene_name "TEST1"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tgene\t100\t110\t.\t-\t.\tgene_name "TEST2";\n' \
                'chr1\tHAVANA\ttranscript\t100\t110\t.\t-\t.\ttranscript_id "ENST_B";gene_name "TEST2"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t100\t110\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t105\t110\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n'\
                'chr2\tHAVANA\tgene\t100\t110\t.\t-\t.\tgene_name "TEST3";\n' \
                'chr2\tHAVANA\ttranscript\t100\t110\t.\t-\t.\ttranscript_id "ENST_C";gene_name "TEST3"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr2\tHAVANA\texon\t100\t110\t.\t-\t.\ttranscript_id "ENST_C" gene_name "TEST3"; transcript_type "protein_coding"\n' \
                'chr2\tHAVANA\tCDS\t105\t110\t.\t-\t.\ttranscript_id "ENST_C" gene_name "TEST3"; transcript_type "protein_coding"\n'
        
        with tempfile.NamedTemporaryFile() as temp, tempfile.NamedTemporaryFile() as fasta:
            write_gtf(temp.name, lines)
            make_fasta(fasta.name, ['chr1', 'chr2'])
            gencode = Gencode(temp.name, fasta.name)
        
        genes = gencode.in_region('chr1', 5, 6)  # shouldn't get anything
        self.assertEqual(genes, [])
        genes = gencode.in_region('chr1', 5, 15)  # should get TEST1 (via start)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1"]))
        genes = gencode.in_region('chr1', 5, 25)  # should get TEST1 (via start, end)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1"]))
        genes = gencode.in_region('chr1', 15, 25)  # should get TEST1 (via end)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1"]))
        genes = gencode.in_region('chr1', 12, 13) # should get TEST1 (via encapsulate)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1"]))
        genes = gencode.in_region('chr1', 5, 105) # should get TEST1 (via start, end), TEST2 (via start)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1", "TEST2"]))
        genes = gencode.in_region('chr1', 5, 115) # should get TEST1 (via start, end), TEST2 (via start, end)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1", "TEST2"]))
        genes = gencode.in_region('chr1', 15, 105) # should get TEST1 (via end), TEST2 (via start)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST1", "TEST2"]))
        genes = gencode.in_region('chr1', 25, 105)  # should get TEST2 (via start)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST2"]))
        genes = gencode.in_region('chr1', 102, 105) # should get TEST2 (via encapsulate)
        self.assertEqual(set(x.symbol for x in genes), set(["TEST2"]))
        genes = gencode.in_region('chr1', 120, 150) # shouldn't get anything
        self.assertEqual(genes, [])
        
        # no overlapping genes right up to the border of a gene
        in_region = gencode.in_region('chr1', 0, 9)
        self.assertEqual(len(in_region), 0)
        in_region = gencode.in_region('chr1', 9, 9) # check a single base before
        self.assertEqual(len(in_region), 0)
        in_region = gencode.in_region('chr1', 21, 30)
        self.assertEqual(len(in_region), 0)
        
        # we find matches if even a single base of the gene is in the region
        in_region = gencode.in_region('chr1', 0, 11)
        self.assertEqual(len(in_region), 1)
    
    def test_gencode_nearest(self):
        ''' test that we can find the nearest gene from Gencode
        '''
        lines = ['##format: gtf\n',
                'chr1\tHAVANA\tgene\t10\t20\t.\t-\t.\tgene_name "TEST1";\n',
                'chr1\tHAVANA\ttranscript\t10\t20\t.\t-\t.\ttranscript_id "ENST_A"; gene_name "TEST1"; transcript_type "protein_coding"; tag "appris_principal_1";\n',
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tgene\t100\t110\t.\t-\t.\tgene_name "TEST2";\n',
                'chr1\tHAVANA\ttranscript\t100\t110\t.\t-\t.\ttranscript_id "ENST_B"; gene_name "TEST2"; transcript_type "protein_coding"; tag "appris_principal_1";\n',
                'chr1\tHAVANA\texon\t100\t110\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tCDS\t105\t110\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n',
                'chr2\tHAVANA\tgene\t100\t110\t.\t-\t.\tgene_name "TEST3";\n',
                'chr2\tHAVANA\ttranscript\t100\t110\t.\t-\t.\ttranscript_id "ENST_C"; gene_name "TEST3"; transcript_type "protein_coding"; tag "appris_principal_1";\n',
                'chr2\tHAVANA\texon\t100\t110\t.\t-\t.\ttranscript_id "ENST_C" gene_name "TEST3"; transcript_type "protein_coding"\n',
                'chr2\tHAVANA\tCDS\t105\t110\t.\t-\t.\ttranscript_id "ENST_C" gene_name "TEST3"; transcript_type "protein_coding"\n']
        with tempfile.NamedTemporaryFile() as temp, tempfile.NamedTemporaryFile() as fasta:
            write_gtf(temp.name, lines)
            make_fasta(fasta.name, ['chr1', 'chr2'])
            gencode = Gencode(temp.name, fasta.name)
            
            self.assertEqual(gencode.nearest('chr1', 5).symbol, 'TEST1')
            self.assertEqual(gencode.nearest('chr1', 10).symbol, 'TEST1')
            self.assertEqual(gencode.nearest('chr1', 11).symbol, 'TEST1')
            self.assertEqual(gencode.nearest('chr1', 21).symbol, 'TEST1')
            self.assertEqual(gencode.nearest('chr1', 80).symbol, 'TEST2')
            self.assertEqual(gencode.nearest('chr1', 105).symbol, 'TEST2')
            self.assertEqual(gencode.nearest('chr1', 2000).symbol, 'TEST2')
        
            # and look for genes on the final chrom, both before and after the final gene
            self.assertEqual(gencode.nearest('chr2', 0).symbol, 'TEST3')
            self.assertEqual(gencode.nearest('chr2', 2000).symbol, 'TEST3')
            
            with self.assertRaises(ValueError):
                gencode.nearest('chrZZZ', 2000)
    
    def test_gencode_canonical(self):
        ''' test we find the correct canonical transcript
        '''
        lines = ['##format: gtf\n',
                'chr1\tHAVANA\tgene\t20\t30\t.\t-\t.\tgene_name "TEST1";\n',
                'chr1\tHAVANA\ttranscript\t20\t30\t.\t-\t.\ttranscript_id "ENST_A"; gene_name "TEST1"; transcript_type "protein_coding"; tag "appris_principal_1";\n',
                'chr1\tHAVANA\texon\t20\t30\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tCDS\t25\t30\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tgene\t100\t110\t.\t-\t.\tgene_name "TEST1";\n',
                'chr1\tHAVANA\ttranscript\t100\t110\t.\t-\t.\ttranscript_id "ENST_B"; gene_name "TEST1"; transcript_type "protein_coding";\n',
                'chr1\tHAVANA\texon\t100\t110\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST1"; transcript_type "protein_coding"\n',
                'chr1\tHAVANA\tCDS\t110\t100\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST1"; transcript_type "protein_coding"\n',
        ]
        with tempfile.NamedTemporaryFile() as temp, tempfile.NamedTemporaryFile() as fasta:
            write_gtf(temp.name, lines)
            make_fasta(fasta.name, ['chr1', 'chr2'])
            gencode = Gencode(temp.name, fasta.name)
        
        gene = gencode['TEST1']
        canonical = gene.canonical 
        self.assertEqual(canonical.get_name(), 'ENST_A')
        
        # give the second transcript the appris_principal tag as well, which
        # given that it has the longer CDS, would make it the canonical now
        lines[-3] = lines[-3].strip() + ' tag "appris_principal_1";\n'
        with tempfile.NamedTemporaryFile() as temp, tempfile.NamedTemporaryFile() as fasta:
            write_gtf(temp.name, lines)
            make_fasta(fasta.name, ['chr1', 'chr2'])
            gencode = Gencode(temp.name, fasta.name)
        
        gene = gencode['TEST1']
        canonical = gene.canonical
        self.assertEqual(canonical.get_name(), 'ENST_B')
    
    def test_parse_gtf_gene_line(self):
        ''' test we can parse a GTF line for a gene feature
        '''
        line = 'chr1\tHAVANA\tgene\t69091\t70008\t.\t+\t.\tgene_id "ENSG00000186092.4";' \
            'gene_type "protein_coding"; gene_status "KNOWN"; gene_name "OR4F5";' \
            'level 2; havana_gene "OTTHUMG00000001094.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'gene', 
            'start': 69091,
            'end': 70008,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'',
            'transcript_type': b'',
            'is_canonical': 0,
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_transcript_line(self):
        ''' test we can parse a GTF line for a transcript feature
        '''
        line = 'chr1\tHAVANA\ttranscript\t69091\t70008\t.\t+\t.\tgene_id "ENSG00000186092.4"; '\
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; ' \
            'gene_status "KNOWN"; gene_name "OR4F5"; transcript_type "protein_coding"; ' \
            'transcript_status "KNOWN"; transcript_name "OR4F5-001"; level 2; ' \
            'protein_id "ENSP00000334393.3"; tag "basic"; transcript_support_level "NA"; ' \
            'tag "appris_principal_1"; tag "CCDS"; ccdsid "CCDS30547.1"; ' \
            'havana_gene "OTTHUMG00000001094.2"; havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'transcript', 
            'start': 69091,
            'end': 70008,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 5,
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_exon_line(self):
        '''test we can parse a GTF line for an exon feature
        '''
        line = 'chr1\tHAVANA\texon\t69091\t70008\t.\t+\t.\tgene_id "ENSG00000186092.4"; ' \
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; ' \
            'gene_status "KNOWN"; gene_name "OR4F5"; transcript_type "protein_coding"; ' \
            'transcript_status "KNOWN"; transcript_name "OR4F5-001"; exon_number 1; ' \
            'exon_id "ENSE00002319515.1"; level 2; protein_id "ENSP00000334393.3"; ' \
            'tag "basic"; transcript_support_level "NA"; tag "appris_principal_1"; ' \
            'tag "CCDS"; ccdsid "CCDS30547.1"; havana_gene "OTTHUMG00000001094.2"; ' \
            'havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'exon', 
            'start': 69091,
            'end': 70008,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 0,  ## exons don't get checked for principal tag
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_cds_line(self):
        '''test we can parse a GTF line for a CDS feature
        '''
        line = 'chr1\tHAVANA\tCDS\t69091\t70005\t.\t+\t0\tgene_id "ENSG00000186092.4"; ' \
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; ' \
            'gene_status "KNOWN"; gene_name "OR4F5"; transcript_type "protein_coding"; ' \
            'transcript_status "KNOWN"; transcript_name "OR4F5-001"; exon_number 1; ' \
            'exon_id "ENSE00002319515.1"; level 2; protein_id "ENSP00000334393.3"; ' \
            'tag "basic"; transcript_support_level "NA"; tag "appris_principal_1"; ' \
            'tag "CCDS"; ccdsid "CCDS30547.1"; havana_gene "OTTHUMG00000001094.2"; ' \
            'havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'CDS', 
            'start': 69091,
            'end': 70005,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 0,  ## CDS don't get checked for principal tag
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_start_codon_line(self):
        '''test we can parse a GTF line for a start codon feature
        '''
        line = 'chr1\tHAVANA\tstart_codon\t69091\t69093\t.\t+\t0\tgene_id "ENSG00000186092.4"; ' \
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; gene_status "KNOWN"; ' \
            'gene_name "OR4F5"; transcript_type "protein_coding"; transcript_status "KNOWN"; ' \
            'transcript_name "OR4F5-001"; exon_number 1; exon_id "ENSE00002319515.1"; level 2; ' \
            'protein_id "ENSP00000334393.3"; tag "basic"; transcript_support_level "NA"; ' \
            'tag "appris_principal_1"; tag "CCDS"; ccdsid "CCDS30547.1"; ' \
            'havana_gene "OTTHUMG00000001094.2"; havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'start_codon', 
            'start': 69091,
            'end': 69093,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 0,
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_stop_codon_line(self):
        '''test we can parse a GTF line for a stop codon feature
        '''
        line = 'chr1\tHAVANA\tstop_codon\t70006\t70008\t.\t+\t0\tgene_id "ENSG00000186092.4"; ' \
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; gene_status "KNOWN";' \
            'gene_name "OR4F5"; transcript_type "protein_coding"; transcript_status "KNOWN"; ' \
            'transcript_name "OR4F5-001"; exon_number 1; exon_id "ENSE00002319515.1"; level 2; ' \
            'protein_id "ENSP00000334393.3"; tag "basic"; transcript_support_level "NA"; ' \
            'tag "appris_principal_1"; tag "CCDS"; ccdsid "CCDS30547.1"; ' \
            'havana_gene "OTTHUMG00000001094.2"; havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'stop_codon', 
            'start': 70006,
            'end': 70008,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 0,
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_UTR_line(self):
        '''test we can parse a GTF line for a UTR feature
        '''
        line = 'chr1\tHAVANA\tUTR\t70006\t70008\t.\t+\t.\tgene_id "ENSG00000186092.4"; ' \
            'transcript_id "ENST00000335137.3"; gene_type "protein_coding"; gene_status "KNOWN"; ' \
            'gene_name "OR4F5"; transcript_type "protein_coding"; transcript_status "KNOWN"; ' \
            'transcript_name "OR4F5-001"; exon_number 1; exon_id "ENSE00002319515.1"; level 2; ' \
            'protein_id "ENSP00000334393.3"; tag "basic"; transcript_support_level "NA"; ' \
            'tag "appris_principal_1"; tag "CCDS"; ccdsid "CCDS30547.1"; ' \
            'havana_gene "OTTHUMG00000001094.2"; havana_transcript "OTTHUMT00000003223.2";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'UTR', 
            'start': 70006,
            'end': 70008,
            'strand': b'+',
            'symbol': b'OR4F5',
            'tx_id': b'ENST00000335137.3',
            'transcript_type': b'protein_coding',
            'is_canonical': 0,
            }
        self.assertEqual(obj, expected)
    
    def test_parse_gtf_minus_strand(self):
        '''test we can parse a GTF line for a feature on the minus strand
        '''
        line = 'chr1\tHAVANA\tCDS\t70006\t70008\t.\t-\t.\tgene_name "TEST";\n'
        obj = _parse_gtfline(line.encode('utf8'))
        expected = {'chrom': b'chr1', 
            'feature': b'CDS', 
            'start': 70006,
            'end': 70008,
            'strand': b'-',
            'symbol': b'TEST',
            'tx_id': b'',
            'transcript_type': b'',
            'is_canonical': 0,
            }
        self.assertEqual(obj, expected)
    
    def test__open_gencode_multi_gene(self):
        '''test we can parse a GTF with multiple genes
        '''
        lines = '##format: gtf\n' \
                'chr1\tHAVANA\tgene\t10\t20\t.\t-\t.\tgene_name "TEST1";\n' \
                'chr1\tHAVANA\ttranscript\t10\t20\t.\t-\t.\ttranscript_id "ENST_A";gene_name "TEST1"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST1"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tgene\t10\t30\t.\t-\t.\tgene_name "TEST2";\n' \
                'chr1\tHAVANA\ttranscript\t10\t30\t.\t-\t.\ttranscript_id "ENST_B";gene_name "TEST2"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t30\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t30\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST2"; transcript_type "protein_coding"\n'
        
        with tempfile.NamedTemporaryFile() as temp:
            write_gtf(temp.name, lines)
            data = _open_gencode(temp.name)
        
        self.assertEqual(len(data), 2)
        symbol1, tx1, is_principal = data[0]
        symbol2, tx2, is_principal = data[1]
        self.assertNotEqual(symbol1, symbol2)
    
    def test__open_gencode_not_coding(self):
        '''test we can parse GTFs without protein coding transcripts
        '''
        lines = '##format: gtf\n' \
                'chr1\tHAVANA\tgene\t10\t20\t.\t-\t.\tgene_name "TEST";\n' \
                'chr1\tHAVANA\ttranscript\t10\t20\t.\t-\t.\ttranscript_id "ENST_A";gene_name "TEST"; transcript_type "processed_transcript"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "processed_transcript"\n' \
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "processed_transcript"\n' \
                        
        with tempfile.NamedTemporaryFile() as temp:
            write_gtf(temp.name, lines)
            data = _open_gencode(temp.name)
        
        self.assertEqual(len(data), 0)
        
        # but if we allow all transcript types (not just protein_coding, we)
        with tempfile.NamedTemporaryFile() as temp:
            write_gtf(temp.name, lines)
            data = _open_gencode(temp.name, coding_only=False)
        
        self.assertEqual(len(data), 1)
        
    def test__open_gencode_multi_transcript(self):
        '''test we can parse a GTF with multiple transcripts for the same gene
        '''
        lines = '##format: gtf\n' \
                'chr1\tHAVANA\tgene\t10\t20\t.\t-\t.\tgene_name "TEST";\n' \
                'chr1\tHAVANA\ttranscript\t10\t20\t.\t-\t.\ttranscript_id "ENST_A";gene_name "TEST"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tgene\t10\t30\t.\t-\t.\tgene_name "TEST";\n' \
                'chr1\tHAVANA\ttranscript\t10\t30\t.\t-\t.\ttranscript_id "ENST_B";gene_name "TEST"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\texon\t10\t30\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t30\t.\t-\t.\ttranscript_id "ENST_B" gene_name "TEST"; transcript_type "protein_coding"\n'
        
        with tempfile.NamedTemporaryFile() as temp:
            write_gtf(temp.name, lines)
            data = _open_gencode(temp.name)
        
        self.assertEqual(len(data), 2)
        symbol1, tx1, is_principal = data[0]
        symbol2, tx2, is_principal = data[1]
        self.assertEqual(symbol1, symbol2)
        self.assertEqual(tx1.get_name(), 'ENST_A')
        self.assertEqual(tx1.get_cds(), [{'start': 15, 'end': 20}])
        self.assertEqual(tx2.get_name(), 'ENST_B')
        self.assertEqual(tx2.get_cds(), [{'start': 15, 'end': 30}])
        
    def test__open_gencode_multi_exon(self):
        '''test we can parse a GTF into transcripts
        '''
        lines = '##format: gtf\n' \
                'chr1\tHAVANA\tgene\t10\t100\t.\t-\t.\tgene_name "TEST";\n' \
                'chr1\tHAVANA\ttranscript\t10\t100\t.\t-\t.\ttranscript_id "ENST_A";gene_name "TEST"; transcript_type "protein_coding"; tag "appris_principal_1";\n' \
                'chr1\tHAVANA\tUTR\t10\t15\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding";\n' \
                'chr1\tHAVANA\texon\t10\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t15\t20\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\texon\t30\t40\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tCDS\t30\t40\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\texon\t90\t100\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n' \
                'chr1\tHAVANA\tUTR\t90\t100\t.\t-\t.\ttranscript_id "ENST_A" gene_name "TEST"; transcript_type "protein_coding"\n'
        
        with tempfile.NamedTemporaryFile() as temp:
            write_gtf(temp.name, lines)
            data = _open_gencode(temp.name)
        
        self.assertEqual(len(data), 1)
        symbol, tx, is_principal = data[0]
        self.assertEqual(symbol, 'TEST')
        self.assertEqual(tx.get_name(), 'ENST_A')
        self.assertEqual(tx.get_strand(), '-')
        self.assertEqual(tx.get_exons(), [{'start': 10, 'end': 20}, {'start': 30, 'end': 40}, {'start': 90, 'end': 100}])
        self.assertEqual(tx.get_cds(), [{'start': 15, 'end': 20}, {'start': 30, 'end': 40}])
    




# cython: language_level=3, boundscheck=False, emit_linenums=True

import bisect
import logging
from pathlib import Path

from libcpp.algorithm cimport lower_bound, upper_bound
from libcpp.vector cimport vector
from libcpp.string cimport string
from libcpp cimport bool
from libcpp.map cimport map

from pyfaidx import Fasta

from denovonear.transcript cimport Tx, Region, CDS_coords
from denovonear.transcript import Transcript

cdef extern from "gtf.h" namespace "gencode":
    cdef struct GTFLine:
        string chrom
        string feature
        int start
        int end
        string strand
        string symbol
        string tx_id
        string transcript_type
        int is_canonical
        
    GTFLine parse_gtfline(string line)

cdef extern from "gencode.h" namespace "gencode":
    cdef struct NamedTx:
        string symbol
        Tx tx
        int is_canonical
    
    cdef struct GenePoint:
        int pos
        string symbol
    
    vector[NamedTx] open_gencode(string, bool)
    bool CompFunc(const GenePoint &l, const GenePoint &r)
    vector[string] _in_region(string chrom, int start, int end, 
        map[string, vector[GenePoint]] & starts, map[string, vector[GenePoint]] & ends,
        int max_window) except+

cpdef _parse_gtfline(string line):
    ''' python function for unit testing GTF parsing
    '''
    return parse_gtfline(line)

cdef _convert_exons(vector[Region] exons):
    ''' convert vector of exon Regions to list of lists
    
    We need exons and CDS as lists of lists for constructing the python 
    Transcript object.
    '''
    return [[y.start, y.end] for y in exons]

cpdef _open_gencode(gtf_path, coding_only=True):
    ''' python function for unit testing loading transcripts from GTF
    '''
    cdef vector[NamedTx] _transcripts = open_gencode(gtf_path.encode('utf8'), coding_only)
    
    transcripts = []
    for x in _transcripts:
        tx = x.tx
        chrom = tx.get_chrom().decode('utf8')
        start = tx.get_start()
        end = tx.get_end()
        exons = _convert_exons(tx.get_exons())
        cds = _convert_exons(tx.get_cds())
        strand = chr(tx.get_strand())
        tx_id = tx.get_name().decode('utf8')
        transcript_type = tx.get_type().decode('utf8')
        transcript = Transcript(tx_id, chrom, start, end, strand, transcript_type, exons, cds, offset=0)
        transcripts.append((x.symbol.decode('utf8'), transcript, x.is_canonical))
    return transcripts

__genome_ = None

cdef class Gene:
    cdef string _symbol
    cdef vector[Tx] _transcripts
    cdef vector[int] _canonical
    cdef str _chrom
    cdef int _start, _end
    def __cinit__(self, symbol):
        if isinstance(symbol, str):
            symbol = symbol.encode('utf8')
        self._symbol = symbol
        self.start = 999999999
        self.end = -999999999
    
    cdef add_tx(self, Tx tx, int is_canonical):
        self._transcripts.push_back(tx)
        self._canonical.push_back(is_canonical)
        self.chrom = tx.get_chrom().decode('utf8')
        self.start = min(self.start, tx.get_start())
        self.end = max(self.end, tx.get_end())
    
    def add_transcript(self, _tx):
        ''' add a Transcript to the gene object
        
        This ends up coping the data from the Transcript object, rather than
        reusing the Tx object contained in the Trnascript, but it's not too much
        time wasted, so long as we don't do this millions of times.
        '''
        assert isinstance(_tx, Transcript)
        # construct a new Tx obect by copying out the relevant data
        cdef string tx_id = _tx.get_name().encode('utf8')
        cdef string chrom = _tx.get_chrom().encode('utf8')
        cdef int start = _tx.get_start()
        cdef int end = _tx.get_end()
        cdef char strand = ord(_tx.get_strand())
        cdef string tx_type = _tx.get_type().encode('utf8')
        cdef vector[vector[int]] exons
        cdef vector[vector[int]] cds
        cdef vector[int] exon
        for x in _tx.get_exons():
            exon = [x['start'], x['end']]
            exons.push_back(exon)
        for x in _tx.get_cds():
            exon = [x['start'], x['end']]
            cds.push_back(exon)
        
        cdef Tx tx = Tx(tx_id, chrom, start, end, strand, tx_type)
        tx.set_exons(exons, cds)
        tx.set_cds(cds)
        
        cdef string seq = _tx.get_genomic_sequence().encode('utf8')
        cdef int offset = _tx.get_genomic_offset()
        if len(seq) > 0:
            if chr(strand) == '-':
                _tx.reverse_complement(seq)
                seq = _tx.reverse_complement(seq).encode('utf8')
            tx.add_genomic_sequence(seq, offset)
        self.add_tx(tx, False)
    
    def __repr__(self):
        chrom = self.chrom
        return f'Gene("{self.symbol}", {chrom}:{self.start}-{self.end})'
    
    @property
    def symbol(self):
        return self._symbol.decode('utf8')
    
    @property
    def chrom(self):
        return self._chrom
    @chrom.setter
    def chrom(self, value):
        self._chrom = value
    
    @property
    def start(self):
        return self._start
    @start.setter
    def start(self, value):
        self._start = value
    
    @property
    def end(self):
        return self._end
    @end.setter
    def end(self, value):
        self._end = value
    
    @property
    def strand(self):
        if self._transcripts.size() > 0:
            return chr(self._transcripts[0].get_strand())
        raise IndexError('no transcripts in gene yet')
    
    cdef _convert_exons(self, vector[Region] exons):
        ''' convert vector of exon Regions to list of lists
        
        We need exons and CDS as lists of lists for constructing the python 
        Transcript object.
        '''
        return [[y.start, y.end] for y in exons]
    
    cdef _to_Transcript(self, Tx tx):
        ''' construct Transcript (python object) from Tx (c++ object)
        '''
        offset = 5 if tx.get_genomic_offset() == 0 else tx.get_genomic_offset()
        chrom = tx.get_chrom().decode('utf8')
        start = tx.get_start()
        end = tx.get_end()
        exons = self._convert_exons(tx.get_exons())
        cds = self._convert_exons(tx.get_cds())
        seq = tx.get_genomic_sequence().decode('utf8')
        if seq == '':
            seq = None
        if seq is None and __genome_ is not None:
            seq = __genome_[chrom][start-1-offset:end-1+offset].seq.upper()
        strand = chr(tx.get_strand())
        if strand == '-' and seq is not None:
            seq = tx.reverse_complement(seq.encode('utf8')).decode('utf8')
        tx_id = tx.get_name().decode('utf8')
        tx_type = tx.get_type().decode('utf8')
        return Transcript(tx_id, chrom, start, end, strand, tx_type, exons, cds, seq, offset=offset)
    
    @property
    def transcripts(self):
        ''' get list of Transcripts for gene, with genomic DNA included
        '''
        return [self._to_Transcript(x) for x in self._transcripts]
    
    cdef int _cds_len(self, Tx tx):
        ''' get length of coding sequence for a Tx object based transcript
        '''
        cdef CDS_coords coords = tx.get_coding_distance(tx.get_cds_end())
        return coords.position + 1
    
    cdef Tx _max_by_cds(self, vector[Tx] transcripts) except *:
        ''' get longest transcript by CDS length
        '''
        cdef Tx max_tx
        length = 0
        for tx in transcripts:
            curr_len = self._cds_len(tx)
            if curr_len > length:
                length = curr_len
                max_tx = tx
        if length == 0:
            raise ValueError('no coding transcripts')
        return max_tx
    
    cdef int _exonic_len(self, Tx tx):
        ''' get length of exonic sequence for a Tx object based transcript
        '''
        length = 0
        for start, end in _convert_exons(tx.get_exons()):
            length += abs(end - start) + 1
        return length
        # return sum(abs(x.end - x.start) + 1 for x in tx.get_exons())
    
    cdef Tx _max_by_exonic(self, vector[Tx] transcripts) except *:
        ''' get longest transcript by CDS length
        '''
        cdef Tx max_tx
        length = 0
        for tx in transcripts:
            curr_len = self._exonic_len(tx)
            if curr_len > length:
                length = curr_len
                max_tx = tx
        if length == 0:
            raise ValueError('no exonic transcripts')
        return max_tx
    
    @property
    def canonical(self):
        ''' find the canonical transcript for a gene.
        
        Canonical is defined as:
            - transcript with Ensembl_canonical tag (peferred)
            - transcript with longest CDS tagged with appris_principal in the GTF
            - if no appris_principal tags for any tx, use the tx with longest CDS
            # TODO: for the last case, maybe check the protein coding subset first
        
        Occasionally there are multiple transcripts tagged as appris_principal
        and with the same longest CDS, we use the first one of those.
        '''
        cdef vector[Tx] canonical
        for i in range(self._transcripts.size()):
            max_score = max(self._canonical)
            if self._canonical[i] == max_score:
                canonical.push_back(self._transcripts[i])
        
        if canonical.size() == 0:
            canonical = self._transcripts
        
        cdef Tx max_tx
        try:
            max_tx = self._max_by_cds(canonical)
        except ValueError:
            max_tx = self._max_by_exonic(canonical)
        
        return self._to_Transcript(max_tx)
    
    def in_any_tx_cds(self, pos):
        ''' find if a pos is in coding region of any transcript of a gene
        '''
        return any(tx.in_coding_region(pos) for tx in self._transcripts)
    
    def distance(self, chrom, pos):
        ''' get distance to nearest boundary of a gene
        '''
        # sanatize the chromosome first
        if self.chrom.startswith('chr') and not chrom.startswith('chr'):
            chrom = f'chr{chrom}'
        elif not self.chrom.startswith('chr') and chrom.startswith('chr'):
            chrom = chrom[3:]
        
        if self.chrom != chrom:
            return None
        if self.start <= pos <= self.end:
            return 0
        return min(abs(self.start - pos), abs(self.end - pos))

cdef class Gencode:
    cdef dict genes
    cdef map[string, vector[GenePoint]] starts, ends
    def __cinit__(self, gencode=None, fasta=None, coding_only=True):
        ''' initialise Gencode
        
        Args:
            gencode: path to gencode annotations file
            fasta: path to fasta for genome matching annotations build
            coding: restrict to protein_coding only by default
        '''
        if gencode is not None and not Path(gencode).exists():
            raise ValueError(f'cannot find gencode at: {gencode}')
        if fasta is not None and not Path(fasta).exists():
            raise ValueError(f'cannot find fasta at: {fasta}')
        self.genes = {}
        if fasta:
            logging.info(f'opening genome fasta: {fasta}')
            global __genome_
            __genome_ = Fasta(str(fasta))
        logging.info(f'opening gencode annotations: {gencode}')
        cdef vector[NamedTx] transcripts
        cdef Gene curr
        if gencode is not None:
            transcripts = open_gencode(str(gencode).encode('utf8'), coding_only)
            for x in transcripts:
                symbol = x.symbol.decode('utf8')
                if symbol not in self.genes:
                    self.genes[symbol] = Gene(symbol.encode('utf8'))
                curr = self.genes[symbol]
                curr.add_tx(x.tx, x.is_canonical)
                self.genes[symbol] = curr
        self._sort()
    
    def _sort(self):
        ''' index by starts and ends, to speed finding genes in a region
        '''
        for symbol in self.genes:
            gene = self.genes[symbol]
            chrom = gene.chrom.encode('utf8')
            symbol = symbol.encode('utf8')
            
            # ensure the chromosome is present
            if self.starts.count(chrom) == 0:
                self.starts[chrom] = []
            if self.ends.count(chrom) == 0:
                self.ends[chrom] = []
            
            self.starts[chrom].push_back(GenePoint(gene.start, symbol))
            self.ends[chrom].push_back(GenePoint(gene.end, symbol))
        
        # sort start and end coords by position
        for x, values in self.starts:
            self.starts[x] = sorted(values, key=lambda x: x['pos'])
        for x, values in self.ends:
            self.ends[x] = sorted(values, key=lambda x: x['pos'])
    
    def __repr__(self):
        return f'Gencode(n_genes={len(self)})'
    def __len__(self):
        return len(self.genes)
    def __getitem__(self, symbol):
        return self.genes[symbol]
    def __iter__(self):
        for x in self.genes:
            yield x
    
    def add_gene(self, gene):
        ''' add another gene to the Gencode object
        '''
        if gene.symbol not in self.genes:
            self.genes[gene.symbol] = gene
        self._sort()
    
    def nearest(self, str chrom, int pos):
        ''' find the nearest gene to a genomic chrom, pos coordinate
        '''
        chrom = f'chr{chrom}' if not chrom.startswith('chr') else chrom
        _chrom = chrom.encode('utf8')
        
        # first, account for any overlapping genespython
        overlaps = self.in_region(chrom, pos-1, pos+1)  # NOTE: possibly fix?
        if len(overlaps) > 0:
            # if we have > 0 prioritise if the position is in the CDS
            cds_overlaps = [x for x in overlaps if x.in_any_tx_cds(pos)]
            if len(cds_overlaps) > 0:
                overlaps = cds_overlaps
            # prioritise the gene with longest CDS (in the canonical tx)
            txs = [x.canonical for x in overlaps]
            lengths = [x.get_coding_distance(x.get_cds_end())['pos'] for x in txs]
            idx = lengths.index(max(lengths))
            return overlaps[idx]
        
        # no overlaps observed, look for the nearest upstream or downstream gene
        cdef GenePoint site = GenePoint(pos, b'A');
        cdef int i = lower_bound(self.starts[_chrom].begin(), self.starts[_chrom].end(), site, CompFunc) - self.starts[_chrom].begin()
        cdef int j = upper_bound(self.ends[_chrom].begin(), self.ends[_chrom].begin(), site, CompFunc) - self.ends[_chrom].begin()
        
        i = min(i, self.starts[_chrom].size() - 1)
        j = min(j, self.starts[_chrom].size() - 1)
        
        downstream = self[self.starts[_chrom][i].symbol.decode('utf8')]
        upstream = self[self.starts[_chrom][j].symbol.decode('utf8')]
        
        if upstream.distance(chrom, pos) <= downstream.distance(chrom, pos):
            return upstream
        else:
            return downstream
    
    def in_region(self, str _chrom, int start, int end, int max_window=2500000):
        ''' find genes within a genomic region
        
        Args:
            chrom: chromosome to search on
            start: start position of region
            end: end position of region
            max_window: some genes encapsulate the region, which means we have 
                to account for gene lengths of up to 2.3 Mb in the human genome.
                This permits extra search space in other organisms.
        
        Returns:
            list of Gene objects
        '''
        symbols = _in_region(_chrom.encode('utf8'), start, end, self.starts,
            self.ends, max_window)
        return [self[x.decode('utf8')] for x in symbols]
    
    def __exit__(self):
        ''' cleanup at exit
        '''
        global __genome_
        __genome_.close()
        __genome_ = None



""" unit test the SequenceMethods class
"""

from __future__ import division
import unittest

from denovonear.transcript import Transcript

class TestSequenceMethodsPy(unittest.TestCase):
    """ unit test the SequenceMethods class
    """
    
    def setUp(self):
        """ construct a Transcript object to add sequence to
        """
        
        self.gene = self.construct_gene()
    
    def construct_gene(self, name='TEST', chrom='1', start=100, end=200,
            strand='+', exons=[(100, 120), (180, 200)],
            cds=[(110, 120), (180, 190)]):
        
        tx = Transcript(name, chrom, start, end, strand)
        tx.set_exons(exons, cds)
        tx.set_cds(cds)
        
        return tx
    
    def test_get_position_on_chrom(self):
        """ test that get_position_on_chrom() works correctly
        """
        
        # check CDS positions for a gene on the forward strand, across two exons
        self.assertEqual(self.gene.get_position_on_chrom(1), 111)
        self.assertEqual(self.gene.get_position_on_chrom(10), 120)
        self.assertEqual(self.gene.get_position_on_chrom(11), 180)
        
        self.assertEqual(self.gene.get_position_on_chrom(10, 1), 121)
        self.assertEqual(self.gene.get_position_on_chrom(11, -1), 179)
        
        # check CDS positions for gene on the reverse strand
        self.gene = self.construct_gene(strand='-')
        self.assertEqual(self.gene.get_position_on_chrom(1), 189)
        self.assertEqual(self.gene.get_position_on_chrom(10), 180)
        self.assertEqual(self.gene.get_position_on_chrom(11), 120)
        
        self.assertEqual(self.gene.get_position_on_chrom(10, 1), 179)
        self.assertEqual(self.gene.get_position_on_chrom(11, -1), 121)
        
        # run through all the CDS positions, and make sure that converting
        # between coordinates gives the same position as original
        for pos in range(self.gene.get_start(), self.gene.get_end()):
            if self.gene.in_coding_region(pos):
                cds = self.gene.get_coding_distance(pos)
                converted = self.gene.get_position_on_chrom(cds['pos'])
                self.assertEqual(pos, converted)
        
        # now try converting for a gene on the plus strand
        self.gene = self.construct_gene()
        for pos in range(self.gene.get_start(), self.gene.get_end()):
            if self.gene.in_coding_region(pos):
                cds = self.gene.get_coding_distance(pos)
                converted = self.gene.get_position_on_chrom(cds['pos'])
                self.assertEqual(pos, converted)
    
    def test_get_codon_number_for_cds_position(self):
        """ test that get_codon_number_for_cds_position() works correctly
        """
        
        self.assertEqual(self.gene.get_codon_number_for_cds_position(0), 0)
        self.assertEqual(self.gene.get_codon_number_for_cds_position(1), 0)
        self.assertEqual(self.gene.get_codon_number_for_cds_position(2), 0)
        self.assertEqual(self.gene.get_codon_number_for_cds_position(3), 1)
        self.assertEqual(self.gene.get_codon_number_for_cds_position(4), 1)
        self.assertEqual(self.gene.get_codon_number_for_cds_position(30000), 10000)
    
    def test_get_position_within_codon(self):
        """ test that get_position_within_codon() works correctly
        """
        
        self.assertEqual(self.gene.get_position_within_codon(0), 0)
        self.assertEqual(self.gene.get_position_within_codon(1), 1)
        self.assertEqual(self.gene.get_position_within_codon(2), 2)
        self.assertEqual(self.gene.get_position_within_codon(3), 0)
    
    def test_add_genomic_sequence(self):
        """ test that add_genomic_sequence() works correctly
        """
        
        self.gene = self.construct_gene(start=0, end=10, exons=[(0, 4), (6, 10)],
            cds=[(2, 4), (6, 8)])
        
        gdna = "AGAGGCCTATT"
        # check that we can add gDNA sequence, even if we haven't got any
        # reference CDS to check
        self.gene.add_genomic_sequence(gdna)
        
        self.gene.add_cds_sequence("AGGCTA")
        
        # check that we can add gDNA after adding cDNA
        self.gene.add_genomic_sequence(gdna)
        self.assertEqual(self.gene.get_cds_sequence(), "AGGCTA")
        
        # now check for a gene on the reverse strand
        gdna = "ATAGGCCTCTT"
        self.gene = self.construct_gene(start=0, end=10, exons=[(0, 4), (6, 10)],
            cds=[(2, 4), (6, 8)], strand='-')
        self.gene.add_cds_sequence("AGGCTC")
        self.gene.add_genomic_sequence(gdna)
        self.assertEqual(self.gene.get_cds_sequence(), "AGGCTC")
        
        # check that we raise an error if the CDS predicted from the gDNA
        # doesn't match the CDS already provided
        self.gene.add_cds_sequence("TTTTTT")
        with self.assertRaises(ValueError):
            self.gene.add_genomic_sequence(gdna)
    
    def test_add_genomic_sequence_expanded(self):
        """ test that add_genomic_sequence() works correctly with extra sequence
        """
        
        # now check when the 5' and 3' sequence is extended beyond the gene
        start = 1
        end = 9
        exons = [(1, 4), (6, 9)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons, cds=cds)
        
        gdna = "AAAGGCCTTT"
        self.gene.add_cds_sequence("AGGCTT")
        
        self.gene.add_genomic_sequence(gdna, offset=1)
        self.assertEqual(self.gene.get_cds_sequence(), "AGGCTT")
    
    def test_get_centered_sequence(self):
        """ test that get_centered_sequence() works correctly
        """
        
        start = 0
        end = 10
        exons = [(0, 4), (6, 10)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons, cds=cds)
        
        gdna = "AAAGGCCTTT"
        self.gene.add_cds_sequence("AGGCTT")
        self.gene.add_genomic_sequence(gdna, offset=0)
        
        # test CDS positions: start, end, and spanning the exon boundaries
        self.assertEqual(self.gene.get_centered_sequence(2), "AAG")
        self.assertEqual(self.gene.get_centered_sequence(3), "AGG")
        self.assertEqual(self.gene.get_centered_sequence(4), "GGC")
        self.assertEqual(self.gene.get_centered_sequence(6), "CCT")
        self.assertEqual(self.gene.get_centered_sequence(7), "CTT")
        self.assertEqual(self.gene.get_centered_sequence(8), "TTT")
        
        # test that positions outside the CDS raise errors
        with self.assertRaises(ValueError):
            self.gene.get_centered_sequence(-1)
        
        with self.assertRaises(ValueError):
            self.gene.get_centered_sequence(10)
    
    def test_get_centered_sequence_minus_strand(self):
        """ test that get_centered_sequence() works correctly on the minus strand
        """
        
        strand = '-'
        start = 0
        end = 10
        exons = [(0, 4), (6, 10)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons,
            cds=cds, strand=strand)
        
        gdna = "AAAGGCCTTTT"
        self.gene.add_cds_sequence("AGGCTT")
        self.gene.add_genomic_sequence(gdna, offset=0)
        
        # test CDS positions: start, end, and spanning the exon boundaries
        self.assertEqual(self.gene.get_centered_sequence(2), "AAA")
        self.assertEqual(self.gene.get_centered_sequence(3), "AAG")
    
    def test_get_centered_sequence_long_region(self):
        """ test that get_centered_sequence() works correctly with longer regions
        """
        
        start = 0
        end = 10
        exons = [(0, 4), (6, 10)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons, cds=cds)
        
        gdna = "AAAGGCCTTT"
        self.gene.add_cds_sequence("AGGCTT")
        self.gene.add_genomic_sequence(gdna, offset=0)
        
        # test CDS positions: start, end, and spanning the exon boundaries
        self.assertEqual(self.gene.get_centered_sequence(2, length=5), "AAAGG")
        
        # check that we raise an error for regions with even numbered lengths
        with self.assertRaises(ValueError):
            self.gene.get_centered_sequence(2, length=4)
    
    def test_get_codon_sequence(self):
        """ test that get_codon_sequence() works correctly
        """
        
        start = 0
        end = 10
        exons = [(0, 4), (6, 10)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons,
            cds=cds)
        
        gdna = "AAAGGCCTTT"
        self.gene.add_cds_sequence("AGGCTT")
        self.gene.add_genomic_sequence(gdna, offset=0)
        
        # self.gene.start = 0
        # self.gene.end = 10
        # self.gene.exons = [(0, 4), (6, 10)]
        # self.gene.cds = [(2, 4), (6, 8)]
        # self.gene.cds_min = 2
        # self.gene.cds_max = 8
        # self.gene.upstream_sequence = "A"
        # self.gene.cds_sequence = "AGGCTT"
        # self.gene.downstream_sequence = "T"
        
        self.assertEqual(self.gene.get_codon_sequence(0), "AGG")
        self.assertEqual(self.gene.get_codon_sequence(1), "CTT")
        
        # check that codon positions outside the CDS region raise errors
        with self.assertRaises(ValueError):
            self.gene.get_codon_sequence(-1)
        with self.assertRaises(ValueError):
            self.gene.get_codon_sequence(3)
    
    def test_get_codon_sequence_minus_strand(self):
        """ test that get_codon_sequence() works correctly
        """
        
        strand = '-'
        start = 0
        end = 10
        exons = [(0, 4), (6, 10)]
        cds = [(2, 4), (6, 8)]
        self.gene = self.construct_gene(start=start, end=end, exons=exons,
            cds=cds, strand=strand)
        
        gdna = "AAAGGCCTTTT"
        self.gene.add_cds_sequence("AGGCTT")
        self.gene.add_genomic_sequence(gdna, offset=0)
        
        self.assertEqual(self.gene.get_codon_sequence(0), "AGG")
        self.assertEqual(self.gene.get_codon_sequence(1), "CTT")
        
        # check that codon positions outside the CDS region raise errors
        with self.assertRaises(ValueError):
            self.gene.get_codon_sequence(-1)
        with self.assertRaises(ValueError):
            self.gene.get_codon_sequence(3)
    
    def test_translate(self):
        """ test that translate() works correctly
        """
        
        # check a couple of codons, including the stop codon
        self.assertEqual(self.gene.translate("AAA"), "K")
        self.assertEqual(self.gene.translate("AAC"), "N")
        self.assertEqual(self.gene.translate("TAG"), "*")
        
        # raise an error for non-IUPAC base containing codons
        with self.assertRaises(ValueError):
            self.gene.translate("FFF")
        
        # can translate multicodon sequences
        self.assertEqual(self.gene.translate("AAAAAC"), "KN")
        
        # raise an error if we translate sequence with an incomplete codon
        with self.assertRaises(ValueError):
            self.assertEqual(self.gene.translate("AA"), "KN")
            self.assertEqual(self.gene.translate("AAAAA"), "KN")
        




import sys
import os
import io
from setuptools import setup
from distutils.core import Extension
from Cython.Build import cythonize

EXTRA_COMPILE_ARGS = ["-std=c++11"]
EXTRA_LINK_ARGS = []

if sys.platform == "darwin":
    EXTRA_COMPILE_ARGS += ["-stdlib=libc++", "-mmacosx-version-min=10.9"]
    EXTRA_LINK_ARGS += ["-stdlib=libc++", "-mmacosx-version-min=10.9"]

weights = cythonize([
    Extension("denovonear.weights",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=["denovonear/weights.pyx",
            "src/weighted_choice.cpp",
            "src/simulate.cpp"],
        include_dirs=["src/"],
        language="c++"),
    Extension("denovonear.transcript",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=[
            "denovonear/transcript.pyx",
            "src/tx.cpp"],
        include_dirs=["src/"],
        language="c++"),
    Extension("denovonear.gencode",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=[
            "denovonear/gencode.pyx",
            "src/gencode.cpp",
            "src/gtf.cpp",
            "src/tx.cpp",
            "src/gzstream/gzstream.C",
        ],
        include_dirs=["src/", "src/gzstream"],
        libraries=["z"],
        language="c++"),
    Extension("denovonear.site_specific_rates",
        extra_compile_args=EXTRA_COMPILE_ARGS,
        extra_link_args=EXTRA_LINK_ARGS,
        sources=[
            "denovonear/site_specific_rates.pyx",
            "src/weighted_choice.cpp",
            "src/tx.cpp",
            "src/site_rates.cpp"],
        include_dirs=["src/"],
        language="c++"),
    ])

setup(name="denovonear",
    description='Package to examine de novo clustering',
    long_description=io.open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    version="0.9.15",
    author="Jeremy McRae",
    author_email="jeremy.mcrae@gmail.com",
    license="MIT",
    url='https://github.com/jeremymcrae/denovonear',
    packages=["denovonear"],
    install_requires=[
        'aiohttp >= 3.0',
        'scipy >= 0.9.0',
        'cython >= 0.19.0',
        'pyfaidx >= 0.5.8',
    ],
    package_data={"denovonear": ['data/rates.txt', 'weights.pxd']},
    entry_points={'console_scripts': ['denovonear = denovonear.__main__:main']},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
    ext_modules=weights,
    test_suite="tests")



""" unit test the WeightedChoice class"""

from __future__ import division

import unittest

from denovonear.weights import WeightedChoice

class TestWeightedChoicePy(unittest.TestCase):
    """ unit test the WeightedChoice class
    """
        
    def test___init__(self):
        """ check that __init__() initiates the object correctly
        """
        
        choices = WeightedChoice()
        
        # check that an object without any possible choices has a cumulative
        # sum of 0, but returns a choice of -1
        self.assertEqual(choices.get_summed_rate(), 0)
        self.assertEqual(choices.choice(), -1)
        
        # check that the type is set correctly
        self.assertEqual(type(choices), WeightedChoice)
        
    def test_add_choice(self):
        """ test that add_choice() works correctly
        """
        
        # check the cumulative sum while adding in new values
        choices = WeightedChoice()
        choices.add_choice(1, 1)
        self.assertEqual(choices.get_summed_rate(), 1)
        choices.add_choice(2, 5)
        self.assertEqual(choices.get_summed_rate(), 6)
        choices.add_choice(3, 10)
        self.assertEqual(choices.get_summed_rate(), 16)
        
        # check that it works for unsorted probabilities
        choices = WeightedChoice()
        choices.add_choice(1, 1)
        choices.add_choice(2, 10)
        choices.add_choice(3, 5)
        self.assertEqual(choices.get_summed_rate(), 16)
        
        # check for very low values, with very high precision (but not
        # necessarily exactly equal)
        choices = WeightedChoice()
        choices.add_choice(1, 5e-9)
        choices.add_choice(2, 1e-8)
        choices.add_choice(3, 1.000000000000005e-10)
        self.assertAlmostEqual(choices.get_summed_rate(), 1.51000000000000005e-8, places=23)
    
    def test_append(self):
        """ test that append() works correctly
        """
        
        # construct two objects
        a = WeightedChoice()
        a.add_choice(1, 0.5)
        
        b = WeightedChoice()
        b.add_choice(2, 1)
        
        # add one object to the other
        a.append(b)
        
        # check that the first object has changed correctly, but the other
        # remains unchanged
        self.assertEqual(a.get_summed_rate(), 1.5)
        self.assertEqual(b.get_summed_rate(), 1.0)


    
    def test_choice(self):
        """ test that choice() works correctly.
        
        Since WeightedChoice is a weighted random sampler, we can't rely on
        getting exact values out, so repeated samples are expected to obtain
        proportions of values equivalent to their weight value. The difference
        to the expected proportion minimises with larger sample sets, but at
        the cost of making the test hang for > 1 second for 1 million samples,
        or > 10 s for 10 million samples.
        """
        
        iterations = 1000000
        
        choices = WeightedChoice()
        choices.add_choice(1, 1)
        choices.add_choice(2, 5)
        s = [ choices.choice() for x in range(iterations) ]
        self.assertAlmostEqual(s.count(1)/len(s), 0.1667, places=2)
        
        # add another choice, then check that all of the choices have been
        # sampled at the expecetd proportions
        choices.add_choice(3, 4)
        s = [ choices.choice() for x in range(iterations) ]
        self.assertAlmostEqual(s.count(1)/len(s), 0.100, places=2)
        self.assertAlmostEqual(s.count(2)/len(s), 0.500, places=2)
        self.assertAlmostEqual(s.count(3)/len(s), 0.400, places=2)
        
        # check that all the choices have been made from the inserted values
        self.assertEqual(set(s), set([1, 2, 3]))
    
    def test_choice_small_numbers(self):
        """ test that choice() works correctly.
        """
        
        iterations = 1000000
        
        # very small numbers at the end still have expected proportions
        choices = WeightedChoice()
        choices.add_choice(1, 1)
        choices.add_choice(2, 5)
        choices.add_choice(3, 0.0001)
        s = [ choices.choice() for x in range(iterations) ]
        self.assertAlmostEqual(s.count(3)/len(s), 0.0001, places=3)
        
        # very small numbers at the start still have expected proportions
        choices = WeightedChoice()
        choices.add_choice(1, 0.0001)
        choices.add_choice(2, 1)
        choices.add_choice(3, 5)
        s = [ choices.choice() for x in range(iterations) ]
        self.assertAlmostEqual(s.count(1)/len(s), 0.0001, places=3)
        
        # check that the sampling works correctly at low weight values
        choices = WeightedChoice()
        
        numbers = range(1000, 3000)
        small = [ x * 0.000000000001 for x in numbers ]
        for (name, prob) in zip(numbers, small):
            choices.add_choice(name, prob)
        
        s = [ choices.choice() for x in range(iterations) ]
        self.assertAlmostEqual(s.count(numbers[0])/len(s), 0.0001, places=3)
    
    def test_choice_with_alleles(self):
        """ test that choice_with_alleles() works correctly.
        """
        
        # if you add a choice with alleles, then check that we get back alleles,
        # and that they are the same
        choices = WeightedChoice()
        choices.add_choice(1, 1, "A", "T")
        self.assertEqual(choices.choice_with_alleles(),
            {'alt': 'T', 'ref': 'A', 'pos': 1, 'offset': 0})
        self.assertEqual(choices.choice(), 1)
        
        # if you add choices without alleles, then default the alleles to "N"
        choices = WeightedChoice()
        choices.add_choice(1, 1)
        self.assertEqual(choices.choice_with_alleles(),
            {'alt': 'N', 'ref': 'N', 'pos': 1, 'offset': 0})
        
        # make sure you can't add multi-base alleles to the choices
        with self.assertRaises(TypeError):
            choices.add_choice(1, 1, "AA", "A")
            choices.add_choice(1, 1, "A", "AG")
        
        # make sure non-zero offsets are returned corectly
        choices = WeightedChoice()
        choices.add_choice(1, 1, "A", "T", 3)
        self.assertEqual(choices.choice_with_alleles(),
            {'alt': 'T', 'ref': 'A', 'pos': 1, 'offset': 3})
        self.assertEqual(choices.choice(), 1)



""" caches sequence information requested from ensembl, so we don't have to
repeatedly re-request information from the REST API if we have done so recently.
"""

import os
import sqlite3
import sys
import time
import random
import zlib
from pathlib import Path
import json
from datetime import datetime
from urllib.parse import urlparse

class EnsemblCache(object):
    """ Instead of repeatedly re-acquiring data from Ensembl each run, cache
    the requested data for faster retrieval
    """
    today = datetime.today()
    def __init__(self, cache_folder):
        """ initialise the class with the local cache folder
        
        Args:
            cache_folder: path to the cache
        """
        cache_folder = Path(cache_folder)
        if not cache_folder.exists():
            cache_folder.mkdir()
        
        # generate a database with tables if it doesn't already exist
        path = cache_folder / "ensembl_cache.db"
        if not path.exists():
            try:
                with sqlite3.connect(str(path)) as conn:
                    with conn as cursor:
                        cursor.execute("CREATE TABLE ensembl " \
                            "(key text PRIMARY KEY, genome_build text, " \
                            "cache_date text, api_version text, data blob)")
            except sqlite3.OperationalError:
                time.sleep(random.uniform(1, 5))
        
        self.conn = sqlite3.connect(str(path))
        self.conn.row_factory = sqlite3.Row
        self.cursor = self.conn.cursor()
    
    def get_cached_data(self, url):
        """ get cached data for a url if stored in the cache and not outdated
        
        Args:
            url: URL for the Ensembl REST service
        
        Returns:
            data if data in cache, else None
        """
        key, build = self.get_key_from_url(url)
        self.cursor.execute("SELECT * FROM ensembl WHERE key=? AND genome_build=?",
            (key, build))
        row = self.cursor.fetchone()
        
        # if the data has been cached, check that it is not out of date
        if row is not None:
            data = zlib.decompress(row["data"])
            diff = self.today - datetime.strptime(row["cache_date"], "%Y-%m-%d")
            if diff.days < 180:
                return data
        
        return None
    
    def cache_url_data(self, url, data, attempt=0):
        """ cache the data retrieved from ensembl
        
        Args:
            url: URL for the Ensembl REST service
            data: response data from Ensembl, in bytes form
        """
        if attempt > 5:
            raise ValueError('too many attempts at writing to the cache')
        
        key, build = self.get_key_from_url(url)
        current_date = datetime.strftime(self.today, "%Y-%m-%d")
        
        compressed = zlib.compress(data)
        t = (key, build, current_date, '9', compressed)
        
        cmd = "INSERT OR REPLACE INTO ensembl " \
            "(key, genome_build, cache_date, api_version, data) VALUES (?,?,?,?,?)"
        try:
            with self.conn as conn:
                conn.execute(cmd, t)
        except sqlite3.OperationalError:
            # if we hit a sqlite locking error, wait a random time so conflicting
            # instances are less likely to reconflict, then retry
            time.sleep(random.uniform(1, 10))
            self.cache_url_data(url, data, attempt + 1)
    
    def get_key_from_url(self, url):
        """ parses the url into a list of folder locations
        
        We take a URL like:
            http://rest.ensembl.org/sequence/id/ENST00000538324?type=genomic;expand_3prime=10
        and turn it into 'sequence.id.ENST00000538324.genomic'
        
        Args:
            url: URL for the Ensembl REST service
        
        Returns:
            a parsed unique database key for the URLs data
        """
        url = urlparse(url)
        path = url.path.strip('/').replace('/', '.')
        
        # find the build from the url, but convert the default server to a build
        build = url.netloc.split('.')[0]
        build = 'grch38' if build == 'rest' else build
        
        # convert "LONG_ID?feature=transcript" to ['LONG_ID', "transcript"] etc
        suffix = url.query.split(';')[0]
        if "=" in suffix:
            _, suffix = suffix.split("=")
        
        key = path + '.' + suffix if suffix != '' else path
        # replace characters not tolerated in keys
        return key.replace(':', '_'), build
